package de.fernunihagen.rtt.ui.testadapter.jquery.internal;

import static de.fernunihagen.rtt.ui.testadapter.jquery.internal.RTTJQueryUIPlugin.getRTTJQueryUIPlugin;

import java.lang.reflect.Method;

import java.util.ArrayList;
import java.util.List;

import org.contract4j5.contract.Pre;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IAdaptable;

import ca.ubc.jquery.api.JQuery;
import ca.ubc.jquery.api.JQueryAPI;
import ca.ubc.jquery.api.JQueryException;
import ca.ubc.jquery.api.JQueryResult;
import ca.ubc.jquery.api.JQueryResultSet;

import de.fernunihagen.rtt.Test;
import de.fernunihagen.rtt.TestParam;
import de.fernunihagen.rtt.TestProcedureStep;

import de.fernunihagen.rtt.TestProcedureStep.FlowDecision;

import de.fernunihagen.rtt.framework.IElementCollector;
import de.fernunihagen.rtt.framework.ITestActor;

import de.fernunihagen.rtt.framework.TestDescription.ParamTypeInfo;

import de.fernunihagen.rtt.framework.testadapter.RefactoringTestDescription;
import de.fernunihagen.rtt.framework.testadapter.RefactoringTestRunState;

import de.fernunihagen.rtt.internal.testadapter.ITestProcedureStepVisitor;
import de.fernunihagen.rtt.internal.testadapter.InitTestProcedureStep;
import de.fernunihagen.rtt.internal.testadapter.NextTestProcedureStep;
import de.fernunihagen.rtt.internal.testadapter.RunTestProcedureStep;
import de.fernunihagen.rtt.internal.testadapter.SkipTestProcedureStep;

import de.fernunihagen.rtt.internal.util.MutableBoolean;

/**
 * JQuery refactoring test run state.
 * 
 * @author Osama El Hosami
 */
public final class JQueryRefactoringTestRunState extends
		RefactoringTestRunState {

	private JQueryResult _current;
	private JQueryResult _pre;
	private JQueryResultSet _currentSet;

	/**
	 * Creates a JQuery refactoring test run context.
	 * 
	 * @param originator
	 *            Originator
	 * @param collector
	 *            Collector
	 * @param description
	 *            Test description
	 */
	@Pre("null != originator && null != collector && null != description")
	public JQueryRefactoringTestRunState(ITestActor originator,
			IElementCollector<? extends IAdaptable> collector,
			RefactoringTestDescription description) {
		super(originator, collector, description);
		_current = null;
		_currentSet = null;
		_pre = null;
	}

	private void reportError(final String message, final Throwable exception)
			throws CoreException {
		getRTTJQueryUIPlugin().reportError(message, exception);
	}

	@Override
	public void reset() {
		disposeResult();
		super.reset();
	}

	/**
	 * Retrieves the query string of test method
	 * 
	 * @param test
	 *            Test method
	 * 
	 * @return Query string
	 */
	public static String getQuery(final Method test) {
		if (!test.isAnnotationPresent(Test.class)) {
			return "";
		}
		return test.getAnnotation(Test.class).value();
	}

	/**
	 * Retrieves the list of chosen vars
	 * 
	 * @param test
	 *            Test method
	 * @param infos
	 *            Parameter type information
	 * 
	 * @return List of chosen vars
	 */
	public static String[] getChosenVars(final Method test,
			final ParamTypeInfo[] infos) {
		List<String> vars = new ArrayList<String>();

		for (ParamTypeInfo info : infos) {
			vars.add(((TestParam) info.annotation).value());
		}
		return vars.toArray(new String[infos.length]);
	}

	/**
	 * Retrieves the current result.
	 * 
	 * @param test
	 *            Test method
	 * @param chosenVars
	 *            List of chosen vars
	 * 
	 * @return Result
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If an error occurs</li>
	 *             </ul>
	 */
	public JQueryResult getResult(final Method test, final String[] chosenVars)
			throws CoreException {
		if (null == _current) {
			try {

				final JQuery query = JQueryAPI.createQuery(getQuery(test));
				query.setChosenVars(chosenVars);

				_currentSet = query.execute();

				seekNextResult();

			} catch (JQueryException e) {
				reportError("Failed to execute query", e);
			}
		}

		return _current;
	}

	private boolean isDuplicate() {
		return null != _pre && null != _current
				&& _pre.toString().equals(_current.toString());
	}

	private boolean seekNextResult() throws CoreException {
		boolean fHasNext = false;

		if (null != _currentSet) {
			try {
				fHasNext = _currentSet.hasNext();
				if (fHasNext) {
					do {
						_pre = _current;
						_current = _currentSet.next();
					} while (isDuplicate()
							&& (fHasNext = _currentSet.hasNext()));
				}
				if (!fHasNext) {
					disposeResult();
				}
			} catch (JQueryException e) {
				fHasNext = false;
				reportError("Failed to seek to next result", e);
			}
		}

		return !fHasNext;
	}

	private void disposeResult() {
		if (null != _currentSet) {
			_currentSet.close();
			_currentSet = null;
			_pre = _current = null;
		}
	}

	@Override
	public void seekNext() throws CoreException {
		if (!isStalled()) {
			disposeResult();
		}
		super.seekNext();
	}

	/**
	 * Does a test procedure step on current result set.
	 * 
	 * @param step
	 *            Test procedure step
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If an error occurs</li>
	 *             </ul>
	 */
	public boolean doTestProcedureStep(TestProcedureStep step)
			throws CoreException {

		if (null == step) {
			step = TestProcedureStep.skip();
		}

		final MutableBoolean fCompleted = new MutableBoolean(true);

		step.accept(new ITestProcedureStepVisitor() {
			@Override
			public void visit(final NextTestProcedureStep step)
					throws CoreException {
				fCompleted.setValue(seekNextResult());
			}

			@Override
			public void visit(SkipTestProcedureStep step) throws CoreException {
				disposeResult();
				fCompleted.setValue(true);
			}

			@Override
			public void visit(RunTestProcedureStep step) throws CoreException {
				final FlowDecision decision = step.getFlowDecision();

				switch (decision) {
				case NEXT:
					fCompleted.setValue(seekNextResult());
					break;
				case STALL:
					fCompleted.setValue(false);
					break;
				case SKIP:
					disposeResult();
					fCompleted.setValue(true);
					break;
				}
			}

			@Override
			public void visit(InitTestProcedureStep step) throws CoreException {
				// nothing to be done
			}
		});

		return fCompleted.getValue();
	}
}
