package de.fernunihagen.rtt.ui.internal.action;

import static de.fernunihagen.rtt.ui.internal.RTTCoreUIPlugin.getRTTCoreUIPlugin;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Logger;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.runtime.CoreException;

import de.fernunihagen.rtt.ui.IActionAdapter;
import de.fernunihagen.rtt.ui.IActionAdapterProvider;

/**
 * Action adapter provider, manages instances of action adapters, which
 * implement the interface {@link IActionAdapter}.
 * 
 * @author Osama El Hosami
 */
@Contract
public final class ActionAdapterProvider implements IActionAdapterProvider {

	@Invar("null != $target")
	private final Map<String, IActionAdapter> _map;
	@Invar("null != $target")
	private String _id;

	private final static String PROP_ID = "component.name";

	private static final class LoggerHolder {
		private static final Logger _DEFAULT = getRTTCoreUIPlugin().getLogger(
				ActionAdapterProvider.class.getName());
	}

	@Post("null != $return")
	private Logger getLogger() {
		return LoggerHolder._DEFAULT;
	}

	/**
	 * Reports a status representing an error message.
	 * 
	 * @param message
	 *            Message or null
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>Always</li>
	 *             </ul>
	 */
	@Pre("null != message")
	protected final void reportError(final String message) throws CoreException {
		getRTTCoreUIPlugin().reportError(message);
	}

	/**
	 * Creates an action adapter provider.
	 */
	public ActionAdapterProvider() {
		_map = Collections
				.synchronizedMap(new HashMap<String, IActionAdapter>());
		_id = getClass().getName().toLowerCase();
	}

	/**
	 * Component activation.
	 * 
	 * @param properties
	 *            Property map
	 */
	protected void activate(final Map<?, ?> properties) {
		_id = (String) properties.get(PROP_ID);

		if (getLogger().isDebugEnabled()) {
			final String message = String.format("[%s (%s) : activated",
					getID(), properties.toString());
			getLogger().debug(message);
		}
	}

	/**
	 * Retrieves the provider identifier.
	 * 
	 * @return Identifier
	 */
	public String getID() {
		return _id;
	}

	/**
	 * Retrieves all registered action adapters.
	 * 
	 * @return All action adapters
	 */
	@Post("null != $return")
	public final IActionAdapter[] get() {
		return _map.values().toArray(new IActionAdapter[0]);
	}

	/**
	 * Adds an action adapter.
	 * 
	 * @param action
	 *            Action adapter
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>On misconfiguration</li>
	 *             </ul>
	 */
	@Pre("null != action")
	public final void addAction(final IActionAdapter action)
			throws CoreException {
		final IActionAdapter oldAction = _map.put(action.getID(), action);
		if (null != oldAction) {
			final String message = String.format(
					"Previously registered action('%s') '%s' has been replaced"
							+ " by '%s'", action.getID(), oldAction.getName(),
					action.getName());
			getLogger().debug(message);
		}
	}

	/**
	 * Removes an action adapter.
	 * 
	 * @param action
	 *            Action adapter
	 */
	@Pre("null != action")
	public final void removeAction(final IActionAdapter action) {
		final String id = action.getID();
		final IActionAdapter removed = _map.remove(id);
		getLogger().assertLog(removed == action,
				"Unrequested actor '" + id + "' has been removed");
	}

	/**
	 * Retrieves an action by identifier.
	 * 
	 * @param id
	 *            Identifier
	 * 
	 * @return Action adapter
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If not found</li>
	 *             </ul>
	 */
	@Pre("null != id")
	@Post("null != $return")
	public final IActionAdapter get(final String id) throws CoreException {
		final IActionAdapter action = _map.get(id);
		if (null == action) {
			reportError("Actor with id '" + id + "' not found");
		}
		return action;
	}
}
