package de.fernunihagen.rtt.ui.internal.jdt;

import static de.fernunihagen.rtt.ui.internal.RTTCoreUIPlugin.getRTTCoreUIPlugin;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.log4j.Logger;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Pre;

import org.eclipse.core.resources.IProject;

import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.SubMonitor;

import org.eclipse.jdt.core.Flags;
import org.eclipse.jdt.core.IClassFile;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IMember;
import org.eclipse.jdt.core.IMethod;
import org.eclipse.jdt.core.IPackageFragmentRoot;
import org.eclipse.jdt.core.IRegion;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.ITypeHierarchy;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jdt.core.JavaModelException;

import org.eclipse.jdt.core.search.IJavaSearchConstants;
import org.eclipse.jdt.core.search.IJavaSearchScope;
import org.eclipse.jdt.core.search.SearchEngine;
import org.eclipse.jdt.core.search.SearchMatch;
import org.eclipse.jdt.core.search.SearchParticipant;
import org.eclipse.jdt.core.search.SearchPattern;
import org.eclipse.jdt.core.search.SearchRequestor;

import de.fernunihagen.rtt.Test;
import de.fernunihagen.rtt.framework.IElementCollector;
import de.fernunihagen.rtt.framework.IElementFilter;

/**
 * Refactoring Tool test collector.
 * 
 * @author Osama El Hosami
 */
@Contract
public class ToolTestCollector implements IElementCollector<IMember> {
	@Invar("null != $target")
	private final List<IType> _list;
	@Invar("null != $target")
	private final IElementFilter<IAdaptable> _filter;
	@Invar("null != $target")
	private final IProject[] _scope;

	/**
	 * Test annotation name
	 */
	public static final String TEST_ANNOTATION = Test.class.getName();

	private static final class LoggerHolder {
		private static final Logger _DEFAULT = getRTTCoreUIPlugin().getLogger(
				ToolTestCollector.class.getName());
	}

	private static Logger getLogger() {
		return LoggerHolder._DEFAULT;
	}

	private static class AnnotationSearchRequestor extends SearchRequestor {

		private final Collection<IType> _result;
		private final ITypeHierarchy _hierarchy;

		public AnnotationSearchRequestor(final ITypeHierarchy hierarchy,
				final Collection<IType> result) {
			_hierarchy = hierarchy;
			_result = result;
		}

		@Override
		public void acceptSearchMatch(final SearchMatch match)
				throws CoreException {
			if (match.getAccuracy() == SearchMatch.A_ACCURATE
					&& !match.isInsideDocComment()) {
				final Object element = match.getElement();

				Assert.isNotNull(element);
				Assert.isTrue(element instanceof IMethod);

				final IMethod method = (IMethod) element;
				addTypeAndSubtypes(method.getDeclaringType());
			}
		}

		private void addTypeAndSubtypes(final IType type) {
			if (_result.add(type)) {
				final IType[] subclasses = _hierarchy.getSubclasses(type);
				for (IType subclass : subclasses) {
					addTypeAndSubtypes(subclass);
				}
			}
		}
	}

	/**
	 * Create a refactoring tool test collector, used to fill given list with
	 * matching elements. List has to be synchronized in case collector is used
	 * concurrently.
	 * 
	 * @param list
	 *            List
	 * @param filter
	 *            Filter
	 * @param scope
	 *            Projects in search scope
	 */
	@Pre("null != list && null != filter && null != scope")
	public ToolTestCollector(final List<IType> list,
			final IElementFilter<IAdaptable> filter, final IProject[] scope) {
		_list = list;
		_filter = filter;
		_scope = scope;
	}

	/**
	 * Creates a refactoring tool test collector, used to collect matching
	 * elements.
	 * 
	 * @param filter
	 *            Filter
	 * @param scope
	 *            Projects in search scope
	 */
	@Pre("null != filter && null != scope")
	public ToolTestCollector(final IElementFilter<IAdaptable> filter,
			final IProject[] scope) {
		this(new ArrayList<IType>(), filter, scope);
	}

	/**
	 * Creates a refactoring tool test collector.
	 * 
	 * @param scope
	 *            Projects in search scope
	 */
	@Pre("null != scope")
	public ToolTestCollector(final IProject[] scope) {
		this(IElementFilter.ALL, scope);
	}

	private static IRegion getRegion(final IJavaElement element)
			throws JavaModelException {
		Assert.isLegal(null != element);

		final IRegion result = JavaCore.newRegion();
		if (element.getElementType() == IJavaElement.JAVA_PROJECT) {
			// for projects only add the contained source folders
			final IPackageFragmentRoot[] roots = ((IJavaProject) element)
					.getPackageFragmentRoots();
			for (int i = 0; i < roots.length; i++) {
				if (!roots[i].isReadOnly() && !roots[i].isExternal()) {
					result.add(roots[i]);
				}
			}
		} else {
			result.add(element);
		}
		return result;
	}

	private static boolean isAccessibleClass(final IType type)
			throws JavaModelException {
		Assert.isLegal(null != type);

		int flags = type.getFlags();
		if (Flags.isInterface(flags)) {
			return false;
		}
		IJavaElement parent = type.getParent();
		while (true) {
			if (parent instanceof ICompilationUnit
					|| parent instanceof IClassFile) {
				return true;
			}
			if (!(parent instanceof IType) || !Flags.isStatic(flags)
					|| !Flags.isPublic(flags)) {
				return false;
			}
			flags = ((IType) parent).getFlags();
			parent = parent.getParent();
		}
	}

	/**
	 * Collects refactoring tool tests.
	 * 
	 * @param element
	 *            Element to analyze
	 * @param filter
	 *            Element filter
	 * @param result
	 *            List to store result
	 * @param pm
	 *            The progress monitor to use for reporting progress to the
	 *            user. It is the caller's responsibility to call done() on the
	 *            given monitor. Accepts <code>null</code>, indicating that no
	 *            progress should be reported and that the operation cannot be
	 *            canceled
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If an error occurs</li>
	 *             </ul>
	 */
	public static void collectTests(final IJavaElement element,
			final IElementFilter<IAdaptable> filter, final List<IType> result,
			final IProgressMonitor pm) throws CoreException {

		Assert.isLegal(null != element);
		Assert.isLegal(null != result);
		Assert.isLegal(null != filter);

		// FIXME: for performance reason, handle IType element separately =>
		// type is a Test?

		try {
			final SubMonitor monitor = SubMonitor.convert(pm,
					"Searching for refactoring tool test(s)", 3);
			getLogger().debug("Searching for refactoring tool test(s)");

			final IRegion region = getRegion(element);
			final ITypeHierarchy hierarchy = JavaCore.newTypeHierarchy(region,
					null, monitor.newChild(1));
			final IType[] allClasses = hierarchy.getAllClasses();

			// search for all types with references to Test, including
			// subclasses. While Test annotations are done on methods
			// duplicates has to be filtered.
			final Set<IType> candidates = new HashSet<IType>(allClasses.length);
			final SearchRequestor requestor = new AnnotationSearchRequestor(
					hierarchy, candidates);

			final IJavaSearchScope scope = SearchEngine.createJavaSearchScope(
					region.getElements(), IJavaSearchScope.SOURCES);
			final int matchRule = SearchPattern.R_EXACT_MATCH
					| SearchPattern.R_CASE_SENSITIVE;
			final SearchPattern testPattern = SearchPattern.createPattern(
					TEST_ANNOTATION, IJavaSearchConstants.ANNOTATION_TYPE,
					IJavaSearchConstants.ANNOTATION_TYPE_REFERENCE, matchRule);

			final SearchParticipant[] searchParticipants = new SearchParticipant[] { SearchEngine
					.getDefaultSearchParticipant() };
			new SearchEngine().search(testPattern, searchParticipants, scope,
					requestor, monitor.newChild(1));

			monitor.setWorkRemaining(candidates.size());

			// collect all matching candidates
			for (IType candidate : candidates) {
				final boolean fAccept = isAccessibleClass(candidate)
						&& !Flags.isAbstract(candidate.getFlags())
						&& filter.accept(candidate);

				if (getLogger().isDebugEnabled()) {
					getLogger().debug(
							candidate.getElementName() + " (" + fAccept + ")");
				}

				if (fAccept) {
					monitor.subTask(candidate.getElementName());
					result.add(candidate);
				}

				monitor.worked(1);
			}

			getLogger().debug("Search done");

		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}

	/**
	 * Finds refactoring tool tests.
	 * 
	 * @param element
	 *            Element to analyze
	 * @param filter
	 *            Element filter
	 * @param pm
	 *            The progress monitor to use for reporting progress to the
	 *            user. It is the caller's responsibility to call done() on the
	 *            given monitor. Accepts <code>null</code>, indicating that no
	 *            progress should be reported and that the operation cannot be
	 *            canceled
	 * 
	 * @return Result
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If an error occurs</li>
	 *             </ul>
	 */
	public static IType[] findTests(final IJavaElement element,
			final IElementFilter<IAdaptable> filter, final IProgressMonitor pm)
			throws CoreException {
		final List<IType> result = new ArrayList<IType>();
		collectTests(element, filter, result, pm);
		return result.toArray(new IType[result.size()]);
	}

	@Override
	public void collect(final IProgressMonitor pm) throws CoreException {
		try {
			clear();

			final IProject[] projects = getScope();

			final SubMonitor monitor = SubMonitor.convert(pm,
					"Collecting refactoring tool test(s)", SubMonitor.UNKNOWN);
			getLogger().debug("Collecting refactoring tool test(s)");

			for (IProject project : projects) {
				project.open(monitor.newChild(1));

				final boolean fIsJavaProject = project
						.hasNature(JavaCore.NATURE_ID);

				if (getLogger().isDebugEnabled()) {
					getLogger().debug(
							project.getName() + " (" + fIsJavaProject + ")");
				}

				if (fIsJavaProject) {
					collectTests(JavaCore.create(project), _filter, _list,
							monitor.newChild(1));
				}
			}

			getLogger().debug("Collection done (" + _list.size() + ")");

		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}

	@Override
	public List<? extends IMember> getMatches() {
		return Collections.unmodifiableList(_list);
	}

	@Override
	public IProject[] getScope() {
		return _scope;
	}

	@Override
	public void clear() {
		_list.clear();
	}

}
