package de.fernunihagen.rtt.ui.internal.jdt.launcher;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.runtime.IAdaptable;

import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IParent;
import org.eclipse.jdt.core.JavaModelException;

import de.fernunihagen.rtt.framework.IElementCollector;
import de.fernunihagen.rtt.framework.ITestActor;

/**
 * Abstract test subject, used as proband for an tool tests.
 * 
 * @author Osama El Hosami
 */
@Contract
public abstract class AbstractTestSubject {
	@Invar("null != $target")
	private final ToolTestConfiguration _configuration;
	@Invar("null != $target")
	private final IJavaElement _adaptee;

	/**
	 * Excluded state
	 */
	public enum ExcludedState {
		IGNORED, NONE, ALL, PARTIAL;

		/**
		 * Combines two states.
		 * 
		 * @param other
		 *            Another state
		 * 
		 * @return new result
		 */
		public ExcludedState combine(final ExcludedState other) {
			if (this == IGNORED) {
				return other;
			}
			if (other == IGNORED) {
				return this;
			}
			if (this != other) {
				return PARTIAL;
			}
			return this;
		}

		/**
		 * Test if state represents an excluded state.
		 * 
		 * @return True if excluded otherwise false
		 */
		public boolean isExcluded() {
			return this == ALL;
		}
	}

	/**
	 * Creates an abstract test subject.
	 * 
	 * @param configuration
	 *            Test configuration
	 * @param element
	 *            Java element
	 */
	@Pre("null != configuration && null != element")
	public AbstractTestSubject(final ToolTestConfiguration configuration,
			final IJavaElement element) {
		_configuration = configuration;
		_adaptee = element;
	}

	@Override
	public int hashCode() {
		return _adaptee.hashCode();
	}

	@Override
	public boolean equals(final Object other) {
		if (other == this) {
			return true;
		}
		if (null == other) {
			return false;
		}
		return other.equals(_adaptee);
	}

	/**
	 * Retrieves the parent element.
	 * 
	 * @return Parent element or null
	 */
	public abstract AbstractTestSubject getParent();

	/**
	 * Retrieves the root element.
	 * 
	 * @return Root element
	 */
	@Post("null != $return")
	public TestSubjectContainer getRoot() {
		return getParent().getRoot();
	}

	/**
	 * Retrieves the associated java element.
	 * 
	 * @return Java element
	 */
	@Post("null != $return")
	public IJavaElement getJavaElement() {
		return _adaptee;
	}

	/**
	 * Retrieves the applicable element collector.
	 * 
	 * @param actor
	 *            Test actor
	 * 
	 * @return Element collector or null
	 */
	@Pre("null != actor")
	protected IElementCollector<? extends IAdaptable> getCollector(
			final ITestActor actor) {
		if (null == getParent()) {
			return null;
		}
		return getRoot().getCollector(actor);
	}

	private final boolean accept(final ITestActor actor,
			boolean fIncludeChildren) {
		boolean fAccept = accept(actor);
		if (!fIncludeChildren || fAccept) {
			return fAccept;
		}

		final AbstractTestSubject[] children = getChildren();
		for (AbstractTestSubject element : children) {
			fAccept = element.accept(actor, fIncludeChildren);
			if (fAccept) {
				break;
			}
		}

		return fAccept;
	}

	/**
	 * Tests if test element is applicable for given actor.
	 * 
	 * @param actor
	 *            Test actor
	 * 
	 * @return True if satisfied otherwise false
	 */
	public final boolean accept(final ITestActor actor) {
		if (null == actor) {
			return false;
		}
		final IElementCollector<? extends IAdaptable> collector = getCollector(actor);
		return null != collector
				&& collector.getMatches().contains(getJavaElement());
	}

	/**
	 * Retrieves the children of tool test element.
	 * 
	 * @return Children
	 */
	@Post("null != $return")
	public abstract AbstractTestSubject[] getChildren();

	/**
	 * Tests if element has children.
	 * 
	 * @return True if element has children
	 */
	public boolean hasChildren() {
		try {
			return _adaptee instanceof IParent
					&& ((IParent) _adaptee).hasChildren();
		} catch (JavaModelException e) {
		}
		return false;
	}

	/**
	 * Excludes element from actor configuration.
	 * 
	 * @param actor
	 *            Test actor
	 * @param fState
	 *            State
	 */
	@Pre("null != actor")
	public void exclude(final ITestActor actor, boolean fState) {
		final AbstractTestSubject[] children = getChildren();
		for (AbstractTestSubject element : children) {
			element.exclude(actor, fState);
		}

		if (!accept(actor, false)) {
			return;
		}

		getConfiguration().exclude(actor, this, fState);
	}

	/**
	 * Notifies changes.
	 */
	void notifyChanges() {
		getConfiguration().setChanged();
		getConfiguration().notifyObservers();
	}

	/**
	 * Tests if element is excluded in test actor configuration.
	 * 
	 * @param actor
	 *            Test actor
	 * 
	 * @return True if satisfied otherwise false
	 */
	public boolean isExcluded(final ITestActor actor) {
		return getExcludedState(actor).isExcluded();
	}

	/**
	 * Retrieves the excluded state of element in test actor configuration.
	 * 
	 * @param actor
	 *            Test actor
	 * 
	 * @return Excluded state
	 */
	public ExcludedState getExcludedState(final ITestActor actor) {
		ExcludedState state;

		if (getConfiguration().isExcluded(actor, this)) {
			state = ExcludedState.ALL;
		} else if (accept(actor)) {
			state = ExcludedState.NONE;
		} else {
			state = ExcludedState.IGNORED;
		}

		final AbstractTestSubject[] children = getChildren();
		for (AbstractTestSubject element : children) {
			state = element.getExcludedState(actor).combine(state);
		}

		return state;
	}

	/**
	 * Tests if element is in sync with repository.
	 * 
	 * @return True if satisfied otherwise false
	 */
	public boolean isInSync() {
		return getConfiguration().isInSync(_adaptee.getResource());
	}

	/**
	 * Retrieves the test configuration.
	 * 
	 * @return Configuration
	 */
	@Post("null != $return")
	public ToolTestConfiguration getConfiguration() {
		return _configuration;
	}
}
