package de.fernunihagen.rtt.ui.internal.jdt.launcher;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.resources.IProject;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.SubMonitor;

import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IPackageFragmentRoot;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jdt.core.JavaModelException;

import de.fernunihagen.rtt.framework.IElementCollector;
import de.fernunihagen.rtt.framework.IElementFilter;
import de.fernunihagen.rtt.framework.ITestActor;

/**
 * Test subject container.
 * 
 * @author Osama El Hosami
 */
@Contract
public final class TestSubjectContainer extends AbstractTestSubject {

	@Invar("null != $target")
	private final Map<ITestActor, IElementCollector<? extends IAdaptable>> _actorElementCache;

	/**
	 * Creates a test subject container.
	 * 
	 * @param configuration
	 *            Test configuration
	 * @param project
	 *            Project element
	 */
	@Pre("null != configuration && null != project")
	public TestSubjectContainer(final ToolTestConfiguration configuration,
			final IProject project) {
		this(configuration, JavaCore.create(project));
	}

	/**
	 * Creates a test subject container.
	 * 
	 * @param configuration
	 *            Test configuration
	 * @param project
	 *            Project element
	 */
	@Pre("null != configuration && null != project")
	public TestSubjectContainer(final ToolTestConfiguration configuration,
			final IJavaProject project) {
		super(configuration, project);
		_actorElementCache = new HashMap<ITestActor, IElementCollector<? extends IAdaptable>>();
	}

	/**
	 * Initializes subject container with selection of applicable element for
	 * given actors.
	 * 
	 * @param actors
	 *            Actors
	 * @param pm
	 *            The progress monitor to use for reporting progress to the
	 *            user. It is the caller's responsibility to call done() on the
	 *            given monitor. Accepts <code>null</code>, indicating that no
	 *            progress should be reported and that the operation cannot be
	 *            canceled
	 */
	public void initialize(final ITestActor[] actors, final IProgressMonitor pm) {
		try {
			final SubMonitor monitor = SubMonitor.convert(pm,
					"Initializing actors targets", actors.length);

			for (ITestActor actor : actors) {
				if (monitor.isCanceled()) {
					break;
				}

				getCollector(actor);
				monitor.worked(1);
			}

		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}

	private AbstractTestSubject[] getFragments(final IJavaProject project) {
		final ArrayList<AbstractTestSubject> fragments = new ArrayList<AbstractTestSubject>();
		try {
			final IPackageFragmentRoot[] children = project
					.getAllPackageFragmentRoots();
			fragments.ensureCapacity(children.length);

			for (IPackageFragmentRoot fragment : children) {
				if (!fragment.isReadOnly() && !fragment.isExternal()) {
					final TestSubject element = new TestSubject(
							getConfiguration(), fragment, this);
					fragments.add(element);
				}
			}
		} catch (JavaModelException e) {
		}
		fragments.trimToSize();
		return fragments.toArray(new AbstractTestSubject[fragments.size()]);
	}

	@Override
	public AbstractTestSubject getParent() {
		return null;
	}

	@Override
	@Post
	public TestSubjectContainer getRoot() {
		return this;
	}

	@Override
	@Pre
	protected IElementCollector<? extends IAdaptable> getCollector(
			final ITestActor actor) {
		IElementCollector<? extends IAdaptable> collector = _actorElementCache
				.get(actor);
		if (null == collector) {
			collector = actor.getCollector(IElementFilter.ALL,
					new IProject[] { getJavaElement().getProject() });

			try {
				collector.collect(null);
				_actorElementCache.put(actor, collector);
			} catch (CoreException e) {
			}
		}

		return collector;
	}

	@Override
	@Post
	public IJavaProject getJavaElement() {
		return (IJavaProject) super.getJavaElement();
	}

	@Override
	@Post
	public AbstractTestSubject[] getChildren() {
		return getFragments(getJavaElement());
	}

}
