package de.fernunihagen.rtt.ui.internal.jdt.testadapter;

import static de.fernunihagen.rtt.ui.internal.RTTCoreUIPlugin.getRTTCoreUIPlugin;

import java.lang.reflect.Method;

import org.apache.log4j.Logger;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.resources.IProject;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.SubMonitor;

import de.fernunihagen.rtt.TestProcedureStep;
import de.fernunihagen.rtt.TestProcedureStep.FlowDecision;

import de.fernunihagen.rtt.framework.IElementCollector;
import de.fernunihagen.rtt.framework.IElementFilter;
import de.fernunihagen.rtt.framework.ITestContext;
import de.fernunihagen.rtt.framework.ITestElement;
import de.fernunihagen.rtt.framework.ITestElementContainer;
import de.fernunihagen.rtt.framework.ITestRunSession;

import de.fernunihagen.rtt.framework.ITestElement.ProgressState;
import de.fernunihagen.rtt.framework.ITestElement.Result;

import de.fernunihagen.rtt.framework.testadapter.IRefactoringStrategy;
import de.fernunihagen.rtt.framework.testadapter.RefactoringTestDescription;
import de.fernunihagen.rtt.framework.testadapter.RefactoringTestElement;
import de.fernunihagen.rtt.framework.testadapter.RefactoringTestRunState;
import de.fernunihagen.rtt.framework.testadapter.ReflectiveRefactoringTestAdapter;
import de.fernunihagen.rtt.framework.testadapter.RefactoringTestRunState.TestAdapter;

import de.fernunihagen.rtt.internal.jdt.ExcludedJavaElementFilter;
import de.fernunihagen.rtt.internal.jdt.JavaProjectCollector;

import de.fernunihagen.rtt.internal.testadapter.RefactoringTestProcedureStep;

import de.fernunihagen.rtt.internal.util.ICommand;

import de.fernunihagen.rtt.ui.internal.util.CrossThreadOperation;

/**
 * Java refactoring test adapter.
 * 
 * @author Osama El Hosami
 */
@Contract
public abstract class AbstractJavaRefactoringTestAdapter extends
		ReflectiveRefactoringTestAdapter {

	private static final class LoggerHolder {
		private static final Logger _DEFAULT = getRTTCoreUIPlugin().getLogger(
				AbstractJavaRefactoringTestAdapter.class.getName());
	}

	@Post("null != $return")
	private Logger getLogger() {
		return LoggerHolder._DEFAULT;
	}

	/**
	 * Create a java refactoring test adapter.
	 */
	public AbstractJavaRefactoringTestAdapter() {
		super();
	}

	private static boolean hasTestProcedureStepReturnType(final Method test) {
		return TestProcedureStep.class.isAssignableFrom(test.getReturnType());
	}

	/**
	 * Adds a structured child element.
	 * 
	 * @param container
	 *            Container
	 * @param label
	 *            Label
	 * @param value
	 *            String
	 */
	@Pre("null != label && null != container && null != value")
	protected void addStructuredChildElement(
			final ITestElementContainer container, final String label,
			final String value) {
		final ITestElementContainer testElementID = container
				.createChildContainer(1, label);
		testElementID.setTestResult(Result.NONE);
		testElementID.createChild(value, Result.NONE);
		testElementID.setProgressState(ProgressState.COMPLETED);
	}

	/**
	 * Adds a structured child element.
	 * 
	 * @param container
	 *            Container
	 * @param label
	 *            Label
	 * @param value
	 *            Integer
	 */
	@Pre("null != label && null != container")
	protected void addStructuredChildElement(
			final ITestElementContainer container, final String label,
			final int value) {
		addStructuredChildElement(container, label, String
				.format("%08X", value));
	}

	/**
	 * Tests if method has no parameters
	 * 
	 * @param test
	 *            Method
	 * 
	 * @return true if method has no parameters
	 */
	public static boolean hasNoParameter(final Method test) {
		return test.getParameterTypes().length == 0;
	}

	@Override
	@Pre
	protected boolean accept(final Method test, final Class<?> testcase) {
		return hasTestProcedureStepReturnType(test);
	}

	/**
	 * Runs a test adapter without parameters.
	 * 
	 * @param testadapter
	 *            Test adapter
	 * @param element
	 *            Element container
	 * @param pm
	 *            The progress monitor to use for reporting progress to the
	 *            user. It is the caller's responsibility to call done() on the
	 *            given monitor. Accepts <code>null</code>, indicating that no
	 *            progress should be reported and that the operation cannot be
	 *            canceled
	 * 
	 * @return true if execution completed otherwise false
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If an error occurs</li>
	 *             </ul>
	 */
	@Pre("null != testadapter && null != element")
	protected final boolean runWithoutParameters(final TestAdapter testadapter,
			final ITestElementContainer element, final IProgressMonitor pm)
			throws CoreException {

		final TestProcedureStep step = (TestProcedureStep) testadapter.invoke();
		return doTestProcedureStep(testadapter.getRunState(), element, step,
				null, pm);
	}

	/**
	 * Formats given test element container with input values to a string
	 * representation.
	 * 
	 * @param inputs
	 *            Test element container
	 * 
	 * @return Textual representation
	 */
	protected String formatInputs(final ITestElementContainer inputs) {
		StringBuffer buffer = new StringBuffer();
		buffer.append("(");
		if (null != inputs) {
			int i = 0;
			for (ITestElement input : inputs.getChildren()) {
				if (i++ != 0) {
					buffer.append(",");
				}
				buffer.append(input.getLabel() + "=");
				if (input instanceof ITestElementContainer
						&& ((ITestElementContainer) input).hasChildren()) {
					final ITestElement element = ((ITestElementContainer) input)
							.getChildren()[0];

					buffer.append((null != element) ? element.getLabel()
							: "null");
				} else {
					buffer.append("null");
				}
			}
		}
		buffer.append(")");
		return buffer.toString();
	}

	/**
	 * Executes the test procedure step.
	 * 
	 * @param state
	 *            Test run state
	 * @param container
	 *            Element container
	 * @param step
	 *            Test procedure step
	 * @param inputs
	 *            Element container with inputs
	 * @param pm
	 *            The progress monitor to use for reporting progress to the
	 *            user. It is the caller's responsibility to call done() on the
	 *            given monitor. Accepts <code>null</code>, indicating that no
	 *            progress should be reported and that the operation cannot be
	 *            canceled
	 * 
	 * @return true if execution completed otherwise false
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If an error occurs</li>
	 *             </ul>
	 */
	@Pre("null != state && null != container")
	protected boolean doTestProcedureStep(final RefactoringTestRunState state,
			final ITestElementContainer container, TestProcedureStep step,
			final ITestElementContainer inputs, final IProgressMonitor pm)
			throws CoreException {

		final ITestElementContainer output = container.createChildContainer(1,
				"Output");

		try {
			final String userDescription = state.getTestAdapter()
					.getDescription();
			final String description = String.format("%s%s", userDescription,
					formatInputs(inputs));

			output.setTarget(description);

			if (null == step) {
				step = TestProcedureStep.skip();
			}

			output.setProgressState(ProgressState.RUNNING);

			if (step instanceof RefactoringTestProcedureStep) {
				final RefactoringTestProcedureStep refactoringStep = (RefactoringTestProcedureStep) step;
				final IRefactoringStrategy strategy = refactoringStep
						.getStrategy();

				if (null != strategy) {
					final RefactoringTestElement element = new RefactoringTestElement(
							output, description);

					output.setTarget(element);

					element.setProgressState(ProgressState.RUNNING);
					performChange(strategy, pm);
					element.setProgressState(ProgressState.COMPLETED);

					state.setStrategy(strategy);
					element.setTestResult(strategy);
				}
				// else Ignored notification
			}
			// else Ignored notification

			output.setProgressState(ProgressState.COMPLETED);

			return step.getFlowDecision() != FlowDecision.STALL;

		} catch (OperationCanceledException e) {
			output.setProgressState(ProgressState.STOPPED);
			throw e;
		} catch (CoreException e) {
			getLogger().trace("Unexpected", e);
			output.setTestResult(Result.ERROR);
			throw e;
		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}

	@Override
	@Pre
	public boolean canUndoChange(final ITestContext context) {
		final RefactoringTestRunState state = getState(context);
		return null != state.getStrategy()
				&& state.getStrategy().canUndoChange();
	}

	@Pre("null != strategy")
	private void undoChange(final IRefactoringStrategy strategy,
			final IProgressMonitor monitor) throws CoreException {
		final CrossThreadOperation<Void> operation = CrossThreadOperation
				.create(new ICommand<Void, CoreException>() {
					@Override
					public Void execute() throws CoreException {
						strategy.undoChange(monitor);
						return null;
					}
				});
		operation.execute();
	}

	@Pre("null != strategy")
	private boolean performChange(final IRefactoringStrategy strategy,
			final IProgressMonitor monitor) throws CoreException {
		final CrossThreadOperation<Boolean> operation = CrossThreadOperation
				.create(new ICommand<Boolean, CoreException>() {
					@Override
					public Boolean execute() throws CoreException {
						return strategy.performChange(monitor);
					}
				});
		return operation.execute();
	}

	@Override
	@Pre
	public void undoChange(final ITestContext context, final IProgressMonitor pm)
			throws CoreException {
		try {
			final RefactoringTestRunState state = getState(context);

			final SubMonitor monitor = SubMonitor.convert(pm,
					"Undoing change(s)", 1);

			final ITestElementContainer container = context
					.createChildContainer(0, "Undo");

			getLogger().debug("Undoing change(s)");

			container.setProgressState(ProgressState.RUNNING);

			if (canUndoChange(context)) {
				undoChange(state.getStrategy(), monitor.newChild(1));
				container.setTestResult(Result.OK);
			} else {
				container.setTestResult(context.getTestResult());
			}

			container.setProgressState(ProgressState.COMPLETED);

			getLogger().debug("Undo done");

		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}

	@Post
	@Pre
	@Override
	public IElementCollector<? extends IAdaptable> getCollector(
			final IElementFilter<IAdaptable> filter, final IProject[] scope) {
		return new JavaProjectCollector(filter, scope);
	}

	@Override
	@Pre
	@Post
	public RefactoringTestRunState create(final ITestRunSession session,
			final IProgressMonitor monitor) throws CoreException {
		final RefactoringTestRunState state;

		try {
			final IProject[] scope = session.getScope();
			final RefactoringTestDescription description = new RefactoringTestDescription(
					session.getConfiguration(this));
			final IElementFilter<IAdaptable> filter = new ExcludedJavaElementFilter(
					description.getExcludedElementIdentifiers());
			state = new RefactoringTestRunState(this, getCollector(filter,
					scope), description);
			state.updateTargets(monitor);

		} finally {
			if (null != monitor) {
				monitor.done();
			}
		}

		return state;
	}
}
