package de.fernunihagen.rtt.ui.internal.util;

import java.io.File;
import java.io.IOException;

import java.lang.reflect.InvocationTargetException;
import java.net.URISyntaxException;
import java.net.URL;

import java.util.List;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IProjectDescription;
import org.eclipse.core.resources.IWorkspace;

import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Path;

import org.eclipse.ui.wizards.datatransfer.ImportOperation;
import org.eclipse.ui.wizards.datatransfer.ZipFileStructureProvider;

import de.fernunihagen.rtt.internal.util.EnumerationIterator;

import de.fernunihagen.rtt.framework.util.IOverwriteQuery;

import static de.fernunihagen.rtt.ui.internal.RTTCoreUIPlugin.getRTTCoreUIPlugin;
import static de.fernunihagen.rtt.internal.ResourceUtil.getResourceUtil;

/**
 * Resource helper.
 * 
 * @author Osama El Hosami
 */
@Contract
public final class ResourceUtil {

	private ResourceUtil() {

	}

	private final static class ResourceUtilHolder {
		private static final ResourceUtil _DEFAULT = new ResourceUtil();
	}

	private static final class OverwriteNoneStrategyHolder {
		private static final IOverwriteQuery _DEFAULT = new IOverwriteQuery() {
			@Override
			public StatusCode queryOverwrite(String pathString) {
				return StatusCode.NO_ALL;
			};
		};
	}

	private static final class OverwriteAllStrategyHolder {
		private static final IOverwriteQuery _DEFAULT = new IOverwriteQuery() {
			@Override
			public StatusCode queryOverwrite(String pathString) {
				return StatusCode.ALL;
			};
		};
	}

	/**
	 * Retrieves the shared overwrite no resource strategy
	 * 
	 * @return The shared overwrite no resource strategy
	 */
	@Post("null != $return")
	public IOverwriteQuery getOverwriteNoneStrategy() {
		return OverwriteNoneStrategyHolder._DEFAULT;
	}

	/**
	 * Retrieves the shared overwrite all resources strategy
	 * 
	 * @return The shared overwrite all resources strategy
	 */
	@Post("null != $return")
	public IOverwriteQuery getOverwriteAllStrategy() {
		return OverwriteAllStrategyHolder._DEFAULT;
	}

	/**
	 * Returns the shared instance
	 * 
	 * @return the shared instance
	 */
	public static ResourceUtil getUIResourceUtil() {
		return ResourceUtilHolder._DEFAULT;
	}

	private void reportError(final String message, final Throwable e)
			throws CoreException {
		getRTTCoreUIPlugin().reportError(message, e);
	}

	private void reportInfo(final String message, final Throwable e)
			throws CoreException {
		getRTTCoreUIPlugin().reportInfo(message, e);
	}

	@Post("null != $return")
	private IWorkspace getWorkspace() {
		return getRTTCoreUIPlugin().getWorkspace();
	}

	@Pre("null != provider && null != fileName")
	private ZipEntry findEntry(final ZipFileStructureProvider provider,
			final String fileName) {
		ZipEntry matching = null;
		final ZipFile archive = provider.getZipFile();
		Assert.isNotNull(archive);

		final Iterable<? extends ZipEntry> entries = EnumerationIterator
				.createIterable(archive.entries());

		for (ZipEntry entry : entries) {
			if (0 == provider.getLabel(entry).compareTo(fileName)) {
				matching = entry;
				break;
			}
		}

		return matching;
	}

	@SuppressWarnings("unchecked")
	@Pre("null != provider && null != entry")
	private List<? extends ZipEntry> getChildren(
			final ZipFileStructureProvider provider, final ZipEntry entry) {
		return provider.getChildren(entry);
	}

	@Pre("null != provider && null != path")
	private ZipEntry findEntry(final ZipFileStructureProvider provider,
			IPath path) {
		ZipEntry matching = provider.getRoot();
		final List<? extends ZipEntry> children = getChildren(provider,
				matching);
		path = path.addTrailingSeparator();

		for (ZipEntry child : children) {
			if (0 == child.getName().compareTo(path.toString())) {
				matching = child;
				break;
			}
		}

		return matching;
	}

	/**
	 * Test if file is a zip archive
	 * 
	 * @param file
	 *            File
	 * 
	 * @return True if satisfied otherwise false
	 */
	public boolean isZipOrJarArchive(final File file) {
		return getResourceUtil().isZipOrJarArchive(file);
	}

	@Pre("null != projectFile && null != query && $this.isZipOrJarArchive(projectFile)")
	private IProject importProjectArchive(final File projectFile,
			final IOverwriteQuery query, final IProgressMonitor monitor)
			throws CoreException {
		ZipFile projectArchive = null;
		try {
			projectArchive = new ZipFile(projectFile);
		} catch (IOException e) {
			reportError("Import failed, while reading project archive", e);
		}

		final ZipFileStructureProvider structureProvider = new ZipFileStructureProvider(
				projectArchive);

		ZipEntry projectDescriptionEntry = findEntry(structureProvider,
				IProjectDescription.DESCRIPTION_FILE_NAME);
		if (null == projectDescriptionEntry) {
			reportError("Import failed, while project archive is invalid ",
					null);
		}

		IProjectDescription projectDescription = null;
		try {
			projectDescription = getWorkspace().loadProjectDescription(
					projectArchive.getInputStream(projectDescriptionEntry));
		} catch (IOException e) {
			reportError("Import failed, while reading project description", e);
		}

		final String name = projectDescription.getName();
		final IPath path = new Path(name);
		final ZipEntry root = findEntry(structureProvider, path);
		Assert.isNotNull(root);

		final ImportOperation op = new ImportOperation(path, root,
				structureProvider,
				new org.eclipse.ui.dialogs.IOverwriteQuery() {
					@Override
					public String queryOverwrite(String pathString) {
						return query.queryOverwrite(pathString).name()
								.toUpperCase();
					}
				});
		op.setCreateContainerStructure(false);
		op.setOverwriteResources(false);

		try {
			op.run(monitor);
		} catch (InvocationTargetException e) {
			reportError("Import failed, while processing operation", e);
		} catch (InterruptedException e) {
			reportInfo("Import canceled, through user request", e);
		}

		return getWorkspace().getRoot().getProject(name);
	}

	private static IProject importProject(final File projectFile,
			final IOverwriteQuery overwriteQuery, final IProgressMonitor monitor) {
		Assert.isLegal(null != projectFile);
		Assert.isLegal(projectFile.isDirectory());

		// TODO handle import of project directory

		return null;
	}

	/**
	 * Imports a project form given project URL
	 * 
	 * @param projectUrl
	 *            Project URL
	 * @param overwriteQuery
	 *            Overwrite strategy or null to use the no overwrite strategy
	 * @param monitor
	 *            The progress monitor to use for reporting progress to the
	 *            user. It is the caller's responsibility to call done() on the
	 *            given monitor. Accepts <code>null</code>, indicating that no
	 *            progress should be reported and that the operation cannot be
	 *            canceled
	 * 
	 * @return Imported project or null
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If an exception occurs during import</li>
	 *             </ul>
	 */
	public IProject importProject(final URL projectUrl,
			IOverwriteQuery overwriteQuery, final IProgressMonitor monitor)
			throws CoreException {
		IProject imported = null;
		File projectFile = null;

		if (null == overwriteQuery) {
			overwriteQuery = getOverwriteNoneStrategy();
		}

		try {
			projectFile = new File(FileLocator.toFileURL(projectUrl).toURI());
		} catch (IOException e) {
			reportError("Import failed, while locating project", e);
		} catch (URISyntaxException e) {
			reportError("Import failed, while locating project", e);
		}

		if (projectFile.exists() && projectFile.canRead()) {
			if (isZipOrJarArchive(projectFile)) {
				imported = importProjectArchive(projectFile, overwriteQuery,
						monitor);
			} else if (projectFile.isDirectory()) {
				imported = importProject(projectFile, overwriteQuery, monitor);
			} else {
				reportError("Import failed, while unsupported project format",
						null);
			}
		} else {
			reportError("Import failed, project inaccessible", null);
		}

		return imported;
	}
}
