package de.fernunihagen.rtt;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.runtime.CoreException;

import org.eclipse.ltk.core.refactoring.Refactoring;

import de.fernunihagen.rtt.framework.testadapter.IRefactoringStrategy;

import de.fernunihagen.rtt.internal.testadapter.ITestProcedureStepVisitor;
import de.fernunihagen.rtt.internal.testadapter.NextTestProcedureStep;
import de.fernunihagen.rtt.internal.testadapter.RefactoringTestProcedureStep;
import de.fernunihagen.rtt.internal.testadapter.SkipTestProcedureStep;

/**
 * Test procedure step. Return value of a test case, combines a test participant
 * with a test control flow decision.
 * 
 * @author Osama El Hosami
 */
@Contract
public abstract class TestProcedureStep {

	public enum FlowDecision {
		NEXT, SKIP, STALL
	}

	private static FlowDecision getFlowDecision(boolean fSeekNext) {
		return (fSeekNext) ? FlowDecision.NEXT : FlowDecision.STALL;
	}

	/**
	 * Creates a test procedure step of given strategy and flow decision.
	 * 
	 * @param strategy
	 *            Refactoring strategy
	 * @param decision
	 *            Control flow decision
	 * 
	 * @return Test procedure step
	 */
	public static TestProcedureStep create(final IRefactoringStrategy strategy,
			final FlowDecision decision) {
		return new RefactoringTestProcedureStep(strategy, decision);
	}

	/**
	 * Creates a test procedure step of given refactoring and flow decision.
	 * 
	 * @param refactoring
	 *            Refactoring
	 * @param decision
	 *            Control flow decision
	 * 
	 * @return Test procedure step
	 */
	public static TestProcedureStep create(final Refactoring refactoring,
			final FlowDecision decision) {
		return new RefactoringTestProcedureStep(refactoring, decision);
	}

	/**
	 * Creates a test procedure step of given strategy and flow decision.
	 * 
	 * @param strategy
	 *            Refactoring strategy
	 * @param fSeekNext
	 *            Flag indicating seek to next test input is requested
	 * 
	 * @return Test procedure step
	 */
	public static TestProcedureStep create(final IRefactoringStrategy strategy,
			final boolean fSeekNext) {
		return create(strategy, getFlowDecision(fSeekNext));
	}

	/**
	 * Creates a test procedure step of given refactoring and flow decision.
	 * 
	 * @param refactoring
	 *            Refactoring
	 * @param fSeekNext
	 *            Flag indicating seek to next test input is requested
	 * 
	 * @return Test procedure step
	 */
	public static TestProcedureStep create(final Refactoring refactoring,
			final boolean fSeekNext) {
		return create(refactoring, getFlowDecision(fSeekNext));
	}

	/**
	 * Creates a test procedure step of given strategy.
	 * 
	 * @param strategy
	 *            Refactoring strategy
	 * 
	 * @return Test procedure step
	 */
	public static TestProcedureStep create(final IRefactoringStrategy strategy) {
		return create(strategy, true);
	}

	/**
	 * Creates a test procedure step of given refactoring.
	 * 
	 * @param refactoring
	 *            Refactoring
	 * 
	 * @return Test procedure step
	 */
	public static TestProcedureStep create(final Refactoring refactoring) {
		return create(refactoring, true);
	}

	/**
	 * Creates a skip test procedure step.
	 * 
	 * @return Test procedure step
	 */
	public static TestProcedureStep skip() {
		return SkipTestProcedureStep.getInstance();
	}

	/**
	 * Creates a next test procedure step.
	 * 
	 * @return Test procedure step
	 */
	public static TestProcedureStep next() {
		return NextTestProcedureStep.getInstance();
	}

	/**
	 * Retrieves the control flow decision.
	 * 
	 * @return Flow decision
	 */
	@Post("null != $return")
	public abstract FlowDecision getFlowDecision();

	/**
	 * Accepts a test procedure step visitor.
	 * 
	 * @param visitor
	 *            Visitor
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If an error occurs</li>
	 *             </ul>
	 */
	@Pre("null != visitor")
	public abstract void accept(ITestProcedureStepVisitor visitor)
			throws CoreException;
}
