package de.fernunihagen.rtt.framework;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.runtime.Assert;

/**
 * Test element interface.
 * 
 * @author Osama El Hosami
 */
@Contract
public interface ITestElement {
	/**
	 * Run stage
	 */
	enum RunStage {
		INIT, RUN, DISPOSE
	}

	/**
	 * Progress state
	 */
	enum ProgressState {
		NOT_STARTED(1), COMPLETED(2), RUNNING(4), STOPPED(8);

		/**
		 * Creates progress state.
		 * 
		 * @param value
		 *            State value
		 */
		@Pre("Integer.bitCount(value) == 1")
		ProgressState(final int value) {
			_value = value;
		}

		/**
		 * Combines two states. The precedence of two states is defined through
		 * the adicity above state constants. For example: <code>
		 * STOPPED.combine(COMPLETED) == STOPPED;
		 * RUNNING.combine(COMPLETED) == RUNNING;
		 * COMPLETED.combine(NOT_STARTED) == COMPLETED;
		 * </code>
		 * 
		 * @param other
		 *            Another state
		 * 
		 * @return new result
		 */
		@Pre("null != other")
		public ProgressState combine(final ProgressState other) {
			final int value = _value | other._value;
			final int index = 31 - Integer.numberOfLeadingZeros(value);
			final ProgressState[] states = ProgressState.values();
			Assert.isTrue(index >= 0 && index < states.length);
			return states[index];
		}

		private final int _value;
	}

	/**
	 * Result
	 */
	enum Result {
		NONE(1), IGNORED(2), OK(4), FAILURE(8), ERROR(16), UNDEFINED(32);

		/**
		 * Creates result.
		 * 
		 * @param value
		 *            Result value
		 */
		@Pre("Integer.bitCount(value) == 1")
		Result(final int value) {
			_value = value;
		}

		/**
		 * Tests if result indicates success
		 * 
		 * @return true if result indicates success
		 */
		public boolean isSuccess() {
			return this == OK || this == IGNORED || this == NONE;
		}

		/**
		 * Combines two results. The precedence of two results is defined
		 * through the adicity of above result constants. For example: <code>
		 * OK.combine(UNDEFINED) == UNDEFINED;
		 * ERROR.combine(FAILURE) == ERROR;
		 * OK.combine(IGNORED) == OK;
		 * </code>
		 * 
		 * @param other
		 *            Another result
		 * 
		 * @return new result
		 */
		@Pre("null != other")
		public Result combine(final Result other) {
			final int value = _value | other._value;
			final int index = 31 - Integer.numberOfLeadingZeros(value);
			final Result[] results = Result.values();
			Assert.isTrue(index >= 0 && index < results.length);
			return results[index];
		}

		private final int _value;
	}

	/**
	 * Retrieves the test target object.
	 * 
	 * @return Test target object.
	 */
	Object getTarget();

	/**
	 * Retrieves the run stage of test element
	 * 
	 * @return Run stage
	 */
	RunStage getStage();

	/**
	 * Retrieve the accumulated test result.
	 * 
	 * @return Test result
	 */
	@Post("null != $return")
	Result getTestResult();

	/**
	 * Retrieves the accumulated progress state.
	 * 
	 * @param includeChilds
	 *            Flag indicating child progress state are included
	 * 
	 * @return Progress state
	 */
	@Post("null != $return")
	ProgressState getProgressState(boolean includeChilds);

	/**
	 * Retrieves the time elapsed in seconds since test started.
	 * 
	 * @return Time elapsed
	 */
	@Post("0 <= $return")
	double getElapsedTimeInSeconds();

	/**
	 * Retrieves the time elapsed in system ticks since test started.
	 * 
	 * @return Time elapsed
	 */
	@Post("0 <= $return")
	long getElapsedTimeInTicks();

	/**
	 * Retrieves the test element container.
	 * 
	 * @return Test element container
	 */
	@Post("null != $return")
	ITestElementContainer getParentContainer();

	/**
	 * Retrieves the test run session.
	 * 
	 * @return Test run session
	 */
	@Post("null != $return")
	ITestRunSession getSession();

	/**
	 * Retrieves the test element label.
	 * 
	 * @return Label
	 */
	@Post("null != $return")
	String getLabel();
}
