package de.fernunihagen.rtt.framework;

import static de.fernunihagen.rtt.internal.RTTCorePlugin.getRTTCorePlugin;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Logger;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.runtime.CoreException;

/**
 * Abstract test actor provider, base class for classes managing instances of
 * test actors, which implement the interface {@link ITestActor}.
 * 
 * @param A
 *            Test actor type
 * 
 * @author Osama El Hosami
 */
@Contract
public abstract class TestActorProvider<A extends ITestActor> implements
		ITestActorProvider<A> {

	@Invar("null != $target")
	private final Map<String, A> _map;
	@Invar("null != $target")
	private String _id;

	private final static String PROP_ID = "component.name";

	private static final class LoggerHolder {
		private static final Logger _DEFAULT = getRTTCorePlugin().getLogger(
				TestActorProvider.class.getName());
	}

	@Post("null != $return")
	private Logger getLogger() {
		return LoggerHolder._DEFAULT;
	}

	/**
	 * Retrieves the immutable map of actors.
	 * 
	 * @return Map of actors
	 */
	@Post("null != $return")
	public final Map<String, A> getMap() {
		return Collections.unmodifiableMap(_map);
	}

	/**
	 * Reports a status representing an error message.
	 * 
	 * @param message
	 *            Message or null
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>Always</li>
	 *             </ul>
	 */
	@Pre("null != message")
	protected final void reportError(final String message) throws CoreException {
		getRTTCorePlugin().reportError(message);
	}

	/**
	 * Creates an abstract actor provider
	 */
	protected TestActorProvider() {
		_map = Collections.synchronizedMap(new HashMap<String, A>());
		_id = getClass().getName().toLowerCase();
	}

	/**
	 * Component activation.
	 * 
	 * @param properties
	 *            Property map
	 */
	protected void activate(final Map<?, ?> properties) {
		_id = (String) properties.get(PROP_ID);

		if (getLogger().isDebugEnabled()) {
			final String message = String.format("[%s (%s) : activated",
					getID(), properties.toString());
			getLogger().debug(message);
		}
	}

	/**
	 * Retrieves the provider identifier.
	 * 
	 * @return Identifier
	 */
	public String getID() {
		return _id;
	}

	/**
	 * Retrieves an empty actor array.
	 * 
	 * @return Empty actor array
	 */
	@Post("null != $return")
	protected abstract A[] getEmptyArray();

	/**
	 * Retrieves all registered actors.
	 * 
	 * @return All actors
	 */
	@Post("null != $return")
	public final A[] get() {
		return _map.values().toArray(getEmptyArray());
	}

	/**
	 * Adds an actor.
	 * 
	 * @param actor
	 *            Actor
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>On misconfiguration</li>
	 *             </ul>
	 */
	@Pre("null != actor")
	public final void addActor(final A actor) throws CoreException {
		final A oldActor = _map.put(actor.getID(), actor);
		if (null != oldActor) {
			final String message = String.format(
					"Previously registered actor('%s') '%s' has been replaced"
							+ " by '%s'", actor.getID(), oldActor.getName(),
					actor.getName());
			getLogger().debug(message);
		}
	}

	/**
	 * Removes an actor.
	 * 
	 * @param actor
	 *            Actor
	 */
	@Pre("null != actor")
	public final void removeActor(final A actor) {
		final String id = actor.getID();
		final A removed = _map.remove(id);
		getLogger().assertLog(removed == actor,
				"Unrequested actor '" + id + "' has been removed");
	}

	/**
	 * Retrieves an actor by identifier.
	 * 
	 * @param id
	 *            Identifier
	 * 
	 * @return Actor
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If not found</li>
	 *             </ul>
	 */
	@Pre("null != id")
	@Post("null != $return")
	public final A get(final String id) throws CoreException {
		final A actor = getMap().get(id);
		if (null == actor) {
			reportError("Actor with id '" + id + "' not found");
		}
		return actor;
	}
}
