package de.fernunihagen.rtt.framework.repository;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;

import de.fernunihagen.rtt.framework.util.IOverwriteQuery;

/**
 * Repository interface.
 * 
 * @author Osama El Hosami
 */
@Contract
public interface IRepository {

	/**
	 * Retrieves the unique repository identifier
	 * 
	 * @return Repository identifier
	 */
	@Post("null != $return")
	String getID();

	/**
	 * Retrieves the repository name
	 * 
	 * @return Repository name
	 */
	@Post("null != $return")
	String getName();

	/**
	 * Return a subscriber that describes the synchronization state of the
	 * resources contained in the projects associated with this repository.
	 * 
	 * @return a subscriber that provides resource synchronization state
	 * 
	 * @throws CoreException
	 */
	@Post("null != $return")
	ISubscriber getSubscriber() throws CoreException;

	/**
	 * For every String in references, load the corresponding project into the
	 * workspace. The opaque strings in references are guaranteed to have been
	 * previously produced by {@link #getReferences(IProject[], IProgressMonitor)}.
	 * The confirmOverwrite method is called with an array of projects for which
	 * projects of the same name already exists in the workspace.
	 * 
	 * @param references
	 *            an array of reference strings uniquely identifying the
	 *            projects
	 * @param query
	 *            Overwrite query instance
	 * @param monitor
	 *            The progress monitor to use for reporting progress to the
	 *            user. It is the caller's responsibility to call done() on the
	 *            given monitor. Accepts <code>null</code>, indicating that no
	 *            progress should be reported and that the operation cannot be
	 *            canceled
	 * 
	 * @return IProject[] an array of projects that were loaded excluding those
	 *         projects already existing and not overwritten
	 * 
	 * @throws CoreException
	 *             thrown if there is a problem loading a project into the
	 *             workspace. If an exception is thrown, then the workspace is
	 *             left in an unspecified state where some of the referenced
	 *             projects may be loaded or partially loaded, and others may
	 *             not.
	 */
	@Pre("null != references && null != query")
	@Post("null != $return")
	IProject[] checkout(String[] references, IOverwriteQuery query,
			IProgressMonitor monitor) throws CoreException;

	/**
	 * For every IProject in projects, return an opaque UTF-8 encoded String to
	 * act as a reference to that project. The format of the String is specific
	 * to the provider. The format of the String must be such that
	 * {@link #checkout(String[], IOverwriteQuery, IProgressMonitor)} will be able to consume it
	 * and recreate a corresponding project.
	 * 
	 * @param projects
	 *            an array of projects that the serializer should create text
	 *            references for
	 * @param monitor
	 *            The progress monitor to use for reporting progress to the
	 *            user. It is the caller's responsibility to call done() on the
	 *            given monitor. Accepts <code>null</code>, indicating that no
	 *            progress should be reported and that the operation cannot be
	 *            canceled
	 * 
	 * @return String[] an array of serialized reference strings uniquely
	 *         identifying the projects
	 * 
	 * @throws CoreException
	 */
	@Pre("null != projects")
	@Post("null != $return")
	String[] getReferences(IProject[] projects, IProgressMonitor monitor)
			throws CoreException;

	/**
	 * Reverts all changes on given resources.
	 * 
	 * @param resources
	 *            Resource list
	 * @param fRecurse
	 *            Flag indicating revert of child elements requested
	 * @param monitor
	 *            The progress monitor to use for reporting progress to the
	 *            user. It is the caller's responsibility to call done() on the
	 *            given monitor. Accepts <code>null</code>, indicating that no
	 *            progress should be reported and that the operation cannot be
	 *            canceled
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If an error occurs</li>
	 *             </ul>
	 */
	@Pre("null != resources")
	void revertChanges(IResource[] resources, boolean fRecurse,
			IProgressMonitor monitor) throws CoreException;

	/**
	 * Refreshes caches and local resources.
	 * 
	 * @param resources
	 *            Resources
	 * @param depth
	 *            Depth
	 * @param monitor
	 *            Progress monitor
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If an error occurs</li>
	 *             </ul>
	 */
	@Pre("null != resources")
	void refreshLocal(IResource[] resources, int depth, IProgressMonitor monitor)
			throws CoreException;
}
