package de.fernunihagen.rtt.framework.repository;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.mapping.ResourceMapping;
import org.eclipse.core.resources.mapping.ResourceTraversal;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;

/**
 * A Subscriber provides synchronization between local resources and a remote
 * location that is used to share those resources.
 * 
 * @author Osama El Hosami
 */
@Contract
public interface ISubscriber {
	/**
	 * Return the name of this subscription, in a format that is suitable for
	 * display to an end user.
	 * 
	 * @return String representing the name of this subscription.
	 */
	@Post("null != $return")
	public String getName();

	/**
	 * Returns <code>true</code> if this resource is supervised by this
	 * subscriber. A supervised resource is one for which this subscriber
	 * maintains the synchronization state. Supervised resources are the only
	 * resources returned when <code>members(IResource)</code> was invoked with
	 * the parent of the resource. Returns <code>false</code> in all other
	 * cases.
	 * 
	 * @param resource
	 *            the resource being tested
	 * @return <code>true</code> if this resource is supervised, and
	 *         <code>false</code> otherwise
	 * @throws CoreException
	 */
	public boolean isSupervised(IResource resource) throws CoreException;

	/**
	 * Returns all non-transient member resources of the given resource. The
	 * result will include entries for resources that exist either in the
	 * workspace or are implicated in an incoming change. Returns an empty list
	 * if the given resource exists neither in the workspace nor in the
	 * corresponding subscriber location, or if the given resource is transient.
	 * <p>
	 * This is a fast operation; the repository is not contacted.
	 * </p>
	 * 
	 * @param resource
	 *            the resource
	 * @return a list of member resources
	 * @throws CoreException
	 */
	@Pre("null != resource")
	@Post("null != $return")
	public IResource[] members(IResource resource) throws CoreException;

	/**
	 * Returns the list of root resources this subscriber considers for
	 * synchronization. A client should call this method first then can safely
	 * call <code>members</code> to navigate the resources managed by this
	 * subscriber.
	 * 
	 * @return a list of resources
	 */
	@Post("null != $return")
	public IResource[] roots();

	/**
	 * Refreshes the resource hierarchy from the given resources and their
	 * children (to the specified depth) from the corresponding resources in the
	 * remote location. Resources are ignored in the following cases:
	 * <ul>
	 * <li>if they do not exist either in the workspace or in the corresponding
	 * remote location</li>
	 * <li>if the given resource is not supervised by this subscriber</li>
	 * <li>if the given resource is a closed project (they are ineligible for
	 * synchronization)</li>
	 * <p>
	 * Typical synchronization operations use the statuses computed by this
	 * method as the basis for determining what to do. It is possible for the
	 * actual sync status of the resource to have changed since the current
	 * local sync status was refreshed. Operations typically skip resources with
	 * stale sync information. The chances of stale information being used can
	 * be reduced by running this method (where feasible) before doing other
	 * operations. Note that this will of course affect performance.
	 * </p>
	 * <p>
	 * The depth parameter controls whether refreshing is performed on just the
	 * given resource (depth= <code>DEPTH_ZERO</code>), the resource and its
	 * children (depth= <code>DEPTH_ONE</code>), or recursively to the resource
	 * and all its descendents (depth= <code>DEPTH_INFINITE</code>). Use depth
	 * <code>DEPTH_ONE</code>, rather than depth <code>DEPTH_ZERO</code>, to
	 * ensure that new members of a project or folder are detected.
	 * </p>
	 * <p>
	 * This method might change resources; any changes will be reported in a
	 * subsequent subscriber resource change event indicating changes to server
	 * sync status.
	 * </p>
	 * <p>
	 * This method contacts the server and is therefore long-running; progress
	 * and cancellation are provided by the given progress monitor.
	 * </p>
	 * 
	 * @param resources
	 *            the resources
	 * @param depth
	 *            valid values are <code>DEPTH_ZERO</code>,
	 *            <code>DEPTH_ONE</code>, or <code>DEPTH_INFINITE</code>
	 * @param monitor
	 *            progress monitor, or <code>null</code> if progress reporting
	 *            and cancellation are not desired
	 * @exception CoreException
	 *                if this method fails. Reasons include:
	 *                <ul>
	 *                <li>The server could not be contacted.</li>
	 *                </ul>
	 */
	@Pre("null != resources")
	public void refresh(IResource[] resources, int depth,
			IProgressMonitor monitor) throws CoreException;

	/**
	 * Adds a listener to this team subscriber. Has no effect if an identical
	 * listener is already registered.
	 * <p>
	 * Resource change listeners are informed about state changes that affect
	 * the resources supervised by this subscriber.
	 * </p>
	 * 
	 * @param listener
	 *            a resource change listener
	 */
	@Pre("null != listener")
	public void addListener(ISubscriberChangeListener listener);

	/**
	 * Removes a listener previously registered with this team subscriber. Has
	 * no affect if an identical listener is not registered.
	 * 
	 * @param listener
	 *            a resource change listener
	 */
	@Pre("null != listener")
	public void removeListener(ISubscriberChangeListener listener);

	/**
	 * Returns synchronization info, in the form of an {@link IDiff} for the
	 * given resource, or <code>null</code> if there is no synchronization info
	 * because the subscriber does not apply to this resource or the resource is
	 * in-sync.
	 * <p>
	 * Note that a diff may be returned for non-existing or for resources which
	 * have no corresponding remote resource.
	 * </p>
	 * <p>
	 * This method will be quick. If synchronization calculation requires
	 * content from the server it must be cached when the subscriber is
	 * refreshed. A client should call refresh before calling this method to
	 * ensure that the latest information is available for computing the diff.
	 * </p>
	 * 
	 * @param resource
	 *            the resource of interest
	 * 
	 * @return the diff for the resource
	 * 
	 * @throws CoreException
	 *             if errors occur
	 */
	@Pre("null != resource")
	@Post("null != $return")
	public IDiff getDiff(IResource resource) throws CoreException;

	/**
	 * Visit any out-of-sync resources covered by the given traversals. Any
	 * resources covered by the traversals are ignored in the following cases:
	 * <ul>
	 * <li>if they do not exist either in the workspace or in the corresponding
	 * remote location</li>
	 * <li>if the given resource is not supervised by this subscriber</li>
	 * <li>if the given resource is a closed project (they are ineligible for
	 * synchronization)</li>
	 * </ul>
	 * 
	 * @param traversals
	 *            the traversals to be visited
	 * @param visitor
	 *            the visitor
	 * @throws CoreException
	 *             if errors occur
	 */
	@Pre("null != traversals && null != visitor")
	public void accept(ResourceTraversal[] traversals, IDiffVisitor visitor)
			throws CoreException;

	/**
	 * Visit any out-of-sync resources in the given resources visited to the
	 * given depth. Resources are ignored in the following cases:
	 * <ul>
	 * <li>if they do not exist either in the workspace or in the corresponding
	 * remote location</li>
	 * <li>if the given resource is not supervised by this subscriber</li>
	 * <li>if the given resource is a closed project (they are ineligible for
	 * synchronization)</li>
	 * </ul>
	 * 
	 * @param resources
	 *            the root of the resource subtrees from which out-of-sync sync
	 *            info should be visited
	 * @param depth
	 *            the depth to which sync info should be collected (one of
	 *            <code>IResource.DEPTH_ZERO</code>,
	 *            <code>IResource.DEPTH_ONE</code>, or
	 *            <code>IResource.DEPTH_INFINITE</code>)
	 * @param visitor
	 *            the visitor
	 * @throws CoreException
	 *             if errors occur
	 */
	@Pre("null != resources && null != visitor")
	public void accept(IResource[] resources, int depth, IDiffVisitor visitor)
			throws CoreException;

	/**
	 * Refresh the subscriber for the given traversals. By default this method
	 * calls {@link #refresh(IResource[], int, IProgressMonitor) } for each
	 * traversal. Any resources covered by the traversals are ignored in the
	 * following cases:
	 * <ul>
	 * <li>if they do not exist either in the workspace or in the corresponding
	 * remote location</li>
	 * <li>if the given resource is not supervised by this subscriber</li>
	 * <li>if the given resource is a closed project (they are ineligible for
	 * synchronization)</li>
	 * </ul>
	 * 
	 * @param traversals
	 *            the traversals to be refreshed
	 * @param monitor
	 *            a progress monitor
	 * @throws CoreException
	 *             if errors occur
	 */
	@Pre("null != traversals")
	public void refresh(ResourceTraversal[] traversals, IProgressMonitor monitor)
			throws CoreException;

	/**
	 * Return the synchronization state of the given resource mapping. Only
	 * return the portion of the synchronization state that matches the provided
	 * stateMask. The synchronization state flags that are guaranteed to be
	 * interpreted by this method are:
	 * <ul>
	 * <li>The kind flags {@link IDiff#ADD}, {@link IDiff#REMOVE} and
	 * {@link IDiff#CHANGE}. If none of these flags are included then all are
	 * assumed.</li>
	 * </ul>
	 * <p>
	 * Other flags can be included and may or may not be interpreted by the
	 * subscriber.
	 * </p>
	 * An element will only include {@link IDiff#ADD} in the returned state if
	 * all resources covered by the traversals mappings are added. Similarly,
	 * {@link IDiff#REMOVE} will only be included if all the resources covered
	 * by the tarversals are deleted. Otherwise {@link IDiff#CHANGE} will be
	 * returned.
	 * 
	 * @param mapping
	 *            the resource mapping whose synchronization state is to be
	 *            determined
	 * @param stateMask
	 *            the mask that identifies the state flags of interested
	 * @param monitor
	 *            a progress monitor
	 * 
	 * @return the synchronization state of the given resource mapping
	 * 
	 * @throws CoreException
	 * 
	 * @see IDiff
	 */
	@Pre("null != mapping")
	public int getState(ResourceMapping mapping, int stateMask,
			IProgressMonitor monitor) throws CoreException;
}
