package de.fernunihagen.rtt.framework.testadapter;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;

import org.eclipse.ltk.core.refactoring.CheckConditionsOperation;
import org.eclipse.ltk.core.refactoring.CreateChangeOperation;
import org.eclipse.ltk.core.refactoring.PerformChangeOperation;
import org.eclipse.ltk.core.refactoring.RefactoringCore;
import org.eclipse.ltk.core.refactoring.RefactoringDescriptorProxy;
import org.eclipse.ltk.core.refactoring.Refactoring;
import org.eclipse.ltk.core.refactoring.RefactoringStatus;

import org.eclipse.ltk.core.refactoring.history.IRefactoringExecutionListener;
import org.eclipse.ltk.core.refactoring.history.IRefactoringHistoryService;
import org.eclipse.ltk.core.refactoring.history.RefactoringExecutionEvent;

/**
 * A Strategy for performing a refactoring. Prior performing a change, all
 * conditions (initial and final) are checked and the change is only performed,
 * if they could be satisfied. In case perform action could be applied, an undo
 * action is provided for the refactoring change, but only if it is supported.
 * 
 * @author Osama El Hosami
 */
@Contract
public final class RefactoringCheckAll implements IRefactoringStrategy,
		IRefactoringExecutionListener {

	private PerformChangeOperation _change;
	@Invar("null != $target")
	private final Refactoring _refactoring;
	private RefactoringDescriptorProxy _proxyDescriptor;

	/**
	 * Creates a refactoring change adapter to encapsulate given refactoring
	 * 
	 * @param refactoring
	 *            A refactoring to apply
	 */
	@Pre("null != refactoring")
	public RefactoringCheckAll(final Refactoring refactoring) {
		_refactoring = refactoring;
		_proxyDescriptor = null;
	}

	@Post("null != $this._change && (!$return || null != $this._proxyDescriptor)")
	@Override
	public boolean performChange(final IProgressMonitor monitor)
			throws CoreException {
		final boolean fExecuted;

		final CheckConditionsOperation check = new CheckConditionsOperation(
				_refactoring, CheckConditionsOperation.ALL_CONDITIONS);

		final CreateChangeOperation create = new CreateChangeOperation(check,
				RefactoringStatus.FATAL);

		_change = new PerformChangeOperation(create);
		_change.setUndoManager(RefactoringCore.getUndoManager(), getName());

		final IRefactoringHistoryService service = RefactoringCore
				.getHistoryService();

		try {
			service.connect();
			service.addExecutionListener(this);

			_change.run(monitor);

		} finally {
			service.removeExecutionListener(this);
			service.disconnect();
		}

		fExecuted = _change.changeExecuted();

		return fExecuted;
	}

	@Override
	public boolean canUndoChange() {
		return _change != null && _change.getUndoChange() != null;
	}

	@Override
	public void undoChange(final IProgressMonitor monitor) throws CoreException {
		if (canUndoChange()) {
			new PerformChangeOperation(_change.getUndoChange()).run(monitor);
			_change = null;
		}
	}

	@Override
	@Post
	public String getName() {
		return _refactoring.getName();
	}

	@Override
	public RefactoringDescriptorProxy getDescriptor() {
		return _proxyDescriptor;
	}

	@Override
	public void executionNotification(RefactoringExecutionEvent event) {
		switch (event.getEventType()) {
		case RefactoringExecutionEvent.PERFORMED:
			_proxyDescriptor = event.getDescriptor();
			break;
		}
	}
}
