package de.fernunihagen.rtt.framework.testadapter;

import static de.fernunihagen.rtt.internal.RTTCorePlugin.getRTTCorePlugin;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;

import java.util.ArrayList;
import java.util.List;
import java.util.NoSuchElementException;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IAdaptable;

import de.fernunihagen.rtt.Test;
import de.fernunihagen.rtt.TestClientSession;
import de.fernunihagen.rtt.TestSession;
import de.fernunihagen.rtt.framework.IElementCollector;
import de.fernunihagen.rtt.framework.ITestActor;
import de.fernunihagen.rtt.framework.ITestRunSession;
import de.fernunihagen.rtt.framework.TestActorRunState;

/**
 * Refactoring test run context.
 * 
 * @author Osama El Hosami
 */
@Contract
public class RefactoringTestRunState extends TestActorRunState {

	@Invar("null != $target")
	private final RefactoringTestDescription _description;
	@Invar("null != $target")
	private final List<TestInfo> _tests;
	@Invar("0 <= $target")
	private int _next;
	@Invar("null != $target")
	private TestAdapter _current;
	@Invar("null != $target")
	private TestAdapter _nextAdapter;
	@Invar("null != $target")
	private final List<CoreException> _errors;
	@Invar("null != $target")
	private IRefactoringStrategy _strategy;
	private boolean _fIsStalled;

	private final static class TestInfo {
		public final Method test;
		public final Class<?> testcase;

		public TestInfo(final Method test, final Class<?> testcase) {
			this.test = test;
			this.testcase = testcase;
		}
	}

	/**
	 * Test adapter.
	 */
	public final class TestAdapter {
		@Invar("null != $target")
		private final TestInfo _info;
		private Object _target;

		private void setFields(final Class<?> clazz, final Object object,
				final Class<? extends Annotation> annotationClass,
				final boolean fIsStatic) throws IllegalArgumentException,
				IllegalAccessException {
			if (null == clazz || null == object) {
				return;
			}

			for (Field field : clazz.getDeclaredFields()) {
				if (field.isAnnotationPresent(annotationClass)) {
					if (!fIsStatic || Modifier.isStatic(field.getModifiers())) {
						final Class<?> fieldClazz = field.getType();
						field.setAccessible(true);
						field.set(_target, fieldClazz.cast(object));
					}
				}
			}

			setFields(clazz.getSuperclass(), object, annotationClass, fIsStatic);
		}

		private void setSessionFields(final Class<?> clazz,
				final TestClientSession session, final boolean fIsStatic)
				throws IllegalArgumentException, IllegalAccessException {
			setFields(clazz, session, TestSession.class, fIsStatic);
		}

		/**
		 * Creates a test adapter.
		 * 
		 * @param session
		 *            Test run session
		 * @param info
		 *            Test info
		 * 
		 * @throws CoreException
		 *             <ul>
		 *             <li>If an error occurs</li>
		 *             </ul>
		 */
		public TestAdapter(final ITestRunSession session, final TestInfo info)
				throws CoreException {
			final boolean fIsStatic = Modifier.isStatic(info.test
					.getModifiers());

			_info = info;

			try {
				_target = (fIsStatic) ? null : info.testcase.newInstance();
				setSessionFields(info.testcase,
						new RefactoringTestClientSession(session,
								RefactoringTestRunState.this), fIsStatic);
			} catch (Exception e) {
				_target = null;
				reportError("Failed to instantiate test case", e);
			}
		}

		/**
		 * Retrieves the test method
		 * 
		 * @return Test method
		 */
		@Post("null != $return")
		public Method getMethod() {
			return _info.test;
		}

		/**
		 * Invokes test adapter
		 * 
		 * @param args
		 *            Arguments
		 * 
		 * @return Return value
		 * 
		 * @throws CoreException
		 *             <ul>
		 *             <li>If an error occurs</li>
		 *             </ul>
		 */
		public Object invoke(final Object... args) throws CoreException {
			try {
				return getMethod().invoke(_target, args);
			} catch (Exception e) {
				reportError("Failed to invoke test", e);
			}
			return null;
		}

		/**
		 * Retrieves the associated run state instance.
		 * 
		 * @return Run state
		 */
		@Post("null != $return")
		public RefactoringTestRunState getRunState() {
			return RefactoringTestRunState.this;
		}

		/**
		 * Retrieves the unique test identifier.
		 * 
		 * @return Unique test identifier
		 */
		public int getID() {
			return getMethod().hashCode();
		}

		/**
		 * Retrieves the unique test instance identifier.
		 * 
		 * @return Unique test instance identifier
		 */
		public int getInstanceID() {
			return System.identityHashCode(_target);
		}

		/**
		 * Retrieves the test case class.
		 * 
		 * @return Test case class
		 */
		@Post("null != $return")
		public Class<?> getTestcase() {
			return _info.testcase;
		}

		/**
		 * Retrieves the textual description of test adapter.
		 * 
		 * @return Description
		 */
		@Post("null != $return")
		public String getDescription() {
			final String description = RefactoringTestRunState
					.getDescription(getMethod());
			if (!description.isEmpty()) {
				return description + "\n" + toString();
			}
			return toString();
		}

		@Override
		@Post("null != $return")
		public String toString() {
			return getMethod().getName();
		}
	}

	/**
	 * Retrieves the textual description of test method.
	 * 
	 * @param test
	 *            Test method
	 * 
	 * @return Description
	 */
	public static String getDescription(final Method test) {
		if (!test.isAnnotationPresent(Test.class)) {
			return "";
		}
		return test.getAnnotation(Test.class).description();
	}

	/**
	 * Creates a refactoring test run context.
	 * 
	 * @param originator
	 *            Originator
	 * @param collector
	 *            Collector
	 * @param description
	 *            Test description
	 */
	@Pre("null != originator && null != collector && null != description")
	public RefactoringTestRunState(final ITestActor originator,
			final IElementCollector<? extends IAdaptable> collector,
			final RefactoringTestDescription description) {
		super(originator, collector);
		_description = description;
		_nextAdapter = _current = null;
		_next = 0;
		_fIsStalled = false;
		_errors = new ArrayList<CoreException>();
		_tests = new ArrayList<TestInfo>();
		_strategy = null;
	}

	private void reportError(final String message, final Throwable exception)
			throws CoreException {
		getRTTCorePlugin().reportError(message, exception);
	}

	private void reportError(final CoreException[] exceptions,
			final String message) throws CoreException {
		getRTTCorePlugin().reportError(exceptions, message);
	}

	@Override
	public void reset() {
		_tests.clear();
		_errors.clear();
		_nextAdapter = _current = null;
		_next = 0;
		_fIsStalled = false;
		_strategy = null;
	}

	@Override
	@Post
	public RefactoringTestDescription getDescription() {
		return _description;
	}

	/**
	 * Adds a test to state object.
	 * 
	 * @param test
	 *            Test
	 * @param testcase
	 *            Test case
	 */
	@Pre("null != test && null != testcase")
	public void addTest(final Method test, final Class<?> testcase) {
		_tests.add(new TestInfo(test, testcase));
	}

	/**
	 * Retrieves the remaining work.
	 * 
	 * @return Work remaining
	 */
	public int getWorkRemaining() {
		return _tests.size() - _next;
	}

	/**
	 * Tests if run has errors.
	 * 
	 * @return true if run generated errors
	 */
	public boolean hasError() {
		return !_errors.isEmpty();
	}

	/**
	 * Reports a status combined from multiple exceptions occurred during run.
	 * 
	 * @param message
	 *            Message
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>Always</li>
	 *             </ul>
	 */
	public void reportError(final String message) throws CoreException {
		final CoreException[] errors = _errors
				.toArray(new CoreException[_errors.size()]);
		_errors.clear();
		reportError(errors, message);
	}

	/**
	 * Adds an exception.
	 * 
	 * @param e
	 *            exception
	 */
	public void addError(final CoreException e) {
		_errors.add(e);
	}

	/**
	 * Retrieves the current test adapter.
	 * 
	 * @return Test adapter
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             If no such element exists
	 *             </ul>
	 */
	@Post("null != $return")
	public TestAdapter getTestAdapter() throws CoreException {
		if (null == _current) {
			reportError(null, new NoSuchElementException());
		}

		return _current;
	}

	/**
	 * Retrieves the next test adapter.
	 * 
	 * @param session
	 *            Test run session
	 * 
	 * @return Test adapter
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             If no such element exists
	 *             </ul>
	 */
	@Post("null != $return")
	public TestAdapter nextTestAdapter(final ITestRunSession session)
			throws CoreException {
		if (!hasNext()) {
			reportError(null, new NoSuchElementException());
		}

		if (!isStalled()) {
			final TestInfo info = _tests.get(_next);
			_nextAdapter = new TestAdapter(session, info);
		}

		Assert.isNotNull(_nextAdapter);
		_current = _nextAdapter;
		return _current;
	}

	/**
	 * Test if run state has more test adapters.
	 * 
	 * @return true if run state has more test adapters
	 */
	public boolean hasNext() {
		return _next >= 0 && _next < _tests.size();
	}

	/**
	 * Tests stall execution state
	 * 
	 * @param fState
	 *            State value
	 */
	public void setStalled(boolean fState) {
		_fIsStalled = fState;
	}

	/**
	 * Tests whether execution is stalled or not.
	 * 
	 * @return true if execution is stalled
	 */
	public boolean isStalled() {
		return _fIsStalled;
	}

	/**
	 * Retrieves seeks to next if execution is not stalled.
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             If no such element exists
	 *             </ul>
	 */
	@Post("null != $return")
	public void seekNext() throws CoreException {
		if (!hasNext()) {
			reportError(null, new NoSuchElementException());
		}
		if (!isStalled()) {
			_next++;
		}
	}

	@Override
	public boolean isActive() {
		return hasNext();
	}

	/**
	 * Sets the active refactoring strategy
	 * 
	 * @param strategy
	 *            Refactoring strategy
	 */
	public void setStrategy(final IRefactoringStrategy strategy) {
		_strategy = strategy;
	}

	/**
	 * Retrieves the active refactoring strategy.
	 * 
	 * @return Refactoring strategy
	 */
	public IRefactoringStrategy getStrategy() {
		return _strategy;
	}
}
