package de.fernunihagen.rtt.framework.util;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

/**
 * Observable interface, used for generic implementation of the Subject-Observer
 * pattern - described in the book "Java Generics and Collections" by
 * <code>Maurice Naftalin</code> and <code>Philip Wadler</code>.
 * 
 * @param <S>
 *            Subject
 * @param <O>
 *            Observer
 * @param <A>
 *            Argument
 * 
 * @author Osama El Hosami
 */
@Contract
public interface IObservable<S extends IObservable<S, O, A>, O extends IObserver<S, O, A>, A> {

	/**
	 * Adds an observer to the set of observers for this object, provided that
	 * it is not the same as some observer already in the set. The order in
	 * which notifications will be delivered to multiple observers is not
	 * specified.
	 * 
	 * @param o
	 *            An observer to be added
	 * 
	 * @throws IllegalArgumentException
	 *             If the parameter o is null
	 */
	@Pre("null != o")
	void addObserver(O o);

	/**
	 * Deletes an observer from the set of observers of this object. Passing
	 * null to this method will have no effect.
	 * 
	 * @param o
	 *            The observer to be deleted
	 */
	@Pre("null != 0")
	void deleteObserver(O o);

	/**
	 * Returns the number of observers of this Observable object.
	 * 
	 * @return The number of observers of this Observable object
	 */
	@Post("0 < $return")
	int countObservers();

	/**
	 * Clears the observer list so that this object no longer has any observers.
	 */
	void deleteObservers();

	/**
	 * Tests if this object has changed.
	 * 
	 * @return True if object has changed otherwise unchanged
	 */
	boolean hasChanged();

	/**
	 * If this object has changed, as indicated by the hasChanged method, then
	 * notify all of its observers and then call the clearChanged method to
	 * indicate that this object has no longer changed.
	 * 
	 * Each observer has its update method called with two arguments: this
	 * observable object and null. In other words, this method is equivalent to:
	 * notifyObservers(null)
	 */
	void notifyObservers();

	/**
	 * If this object has changed, as indicated by the hasChanged method, then
	 * notify all of its observers and then call the clearChanged method to
	 * indicate that this object has no longer changed.
	 * 
	 * Each observer has its update method called with two arguments: this
	 * observable object and the arg argument.
	 * 
	 * @param arg
	 *            Any object
	 */
	void notifyObservers(A arg);
}
