package de.fernunihagen.rtt.framework.util;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

/**
 * Property container interface.
 * 
 * @author Osama El Hosami
 * 
 * @param <S>
 *            Property subject type
 */
@Contract
public interface IPropertyContainer<S>
		extends
		IObservable<IPropertyContainer<S>, IPropertyChangeListener<S>, PropertyChangeEvent> {

	/**
	 * Clear container. 
	 */
	void clear();
	
	/**
	 * Retrieves the property container subject.
	 * 
	 * @return Subject or null if not applicable
	 */
	S getSubject();

	/**
	 * Returns the property value associated with the given key.
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @return Property value associated with the given key
	 */
	@Pre("null != key")
	Object getProperty(String key);

	/**
	 * Returns the property value associated with the given key, or a default
	 * value.
	 * 
	 * @param key
	 *            Property key
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	@Pre("null != key")
	@Post("null == defaultValue || null != $return")
	Object getProperty(String key, Object defaultValue);

	/**
	 * Returns all existing property keys, or an empty array if this property
	 * container has no properties.
	 * 
	 * @return All property keys contained
	 */
	@Post("null != $return")
	String[] getPropertyKeys();

	/**
	 * Returns true if this property container has a property accessible through
	 * the given key, false otherwise.
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @return True if this property container has a property accessible through
	 *         the given key, false otherwise
	 */
	boolean hasProperty(String key);

	/**
	 * Removes the property associated with the given key and returns the old
	 * value. If there's no property associated with the key, null will be
	 * returned.
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @return Property value that used to be associated with the given key
	 */
	@Pre("null != key")
	Object removeProperty(String key);

	/**
	 * Sets the property value for the given key to value.
	 * 
	 * @param key
	 *            Key with which the new property value will be associated
	 * @param value
	 *            Property value
	 */
	@Pre("null != key && null != value")
	void setProperty(String key, Object value);
}
