package de.fernunihagen.rtt.internal;

import static de.fernunihagen.rtt.internal.ResourceUtil.getResourceUtil;

import java.io.IOException;
import java.io.InputStream;

import java.net.URL;

import java.util.Properties;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Plugin;
import org.eclipse.core.runtime.Status;

import com.tools.logging.PluginLogManager;


/**
 * A log manager for plugin's.
 * 
 * @author Osama El Hosami
 */
@Contract
public final class LogManager extends AbstractLogManager {
	private volatile PluginLogManager _manager;
	@Invar("null != $target")
	private final PropertyLocation _properties;

	@Pre("null != plugin && null != properties")
	private LogManager(final Plugin plugin, final PropertyLocation properties) {
		super(plugin);
		_properties = properties;
		_manager = null;
	}

	@Contract
	private abstract static class PropertyLocation {
		private PropertyLocation() {
		}

		/**
		 * Retrieves the property location as URL.
		 * 
		 * @return Property location as URL
		 */
		@Post("null != $return")
		public abstract URL toURL();
	}

	@Contract
	private final static class GeneralPropertyLocation extends PropertyLocation {
		@Invar("null != $target")
		private final URL _properties;

		@Pre("null != properties")
		private GeneralPropertyLocation(final URL properties) {
			super();
			_properties = properties;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see de.fernunihagen.rtt.internal.util.LogManager.PropertyLocation
		 * #toURL()
		 */
		@Override
		@Post()
		public URL toURL() {
			return _properties;
		}

	}

	@Contract
	private final static class BundlePropertyLocation extends PropertyLocation {
		@Invar("null != $target")
		private final String _properties;
		@Invar("null != $target")
		private final Plugin _plugin;

		@Pre("null != plugin && null != properties")
		private BundlePropertyLocation(final Plugin plugin,
				final String properties) {
			super();
			_properties = properties;
			_plugin = plugin;
		}

		@Override
		@Post()
		public URL toURL() {
			return _plugin.getBundle().getEntry("/" + _properties);
		}
	}

	/**
	 * Creates log manager using given property file URL on initialization.
	 * 
	 * @param plugin
	 *            Plug-in
	 * @param properties
	 *            Property file URL
	 */
	@Pre("null != plugin && null != properties")
	public LogManager(final Plugin plugin, final URL properties) {
		this(plugin, new GeneralPropertyLocation(properties));
	}

	/**
	 * Creates log manager using given property bundle file URL on
	 * initialization
	 * 
	 * @param plugin
	 *            Plug-in
	 * @param properties
	 *            Property file URL, contained in bundle
	 */
	@Pre("null != plugin && null != properties")
	public LogManager(final Plugin plugin, final String properties) {
		this(plugin, new BundlePropertyLocation(plugin, properties));
	}

	@Override
	@Post
	protected PluginLogManager getAdaptee() {
		init();
		return _manager;
	}

	/**
	 * Tests if log manager has been closed
	 * 
	 * @return True if satisfied otherwise false
	 */
	@Override
	public boolean isClosed() {
		return null == _manager;
	}

	@Override
	@Post("$return == !$this.isClosed()")
	public boolean init() {
		// double checked locking - should work in JVM's greater or equal to
		// version 1.5
		boolean fInit = !isClosed();
		if (!fInit) {
			synchronized (this) {
				fInit = !isClosed();
				if (!fInit) {
					final String id = getOwnerId();
					InputStream propertiesInputStream = null;

					try {
						propertiesInputStream = _properties.toURL()
								.openStream();
						if (null != propertiesInputStream) {
							final Properties props = new Properties();
							props.load(propertiesInputStream);
							_manager = new PluginLogManager(getOwner(), props);
							fInit = _manager
									.hookPlugin(id, getOwner().getLog());
						}
					} catch (IOException e) {
						final String message = "Error while initializing log properties.";
						final IStatus status = new Status(IStatus.ERROR, id,
								message, e);
						getOwner().getLog().log(status);
					} finally {
						getResourceUtil().close(propertiesInputStream);
					}
				}
			}
		}
		return fInit;
	}

	@Override
	@Post("$this.isClosed()")
	public void close() throws IOException {
		synchronized (this) {
			if (!isClosed()) {
				unhookPlugin(getOwner());
				_manager.shutdown();
				_manager = null;
			}
		}
	}

}
