package de.fernunihagen.rtt.internal;

import java.io.Closeable;
import java.io.IOException;

import org.apache.log4j.Logger;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.resources.IWorkspace;
import org.eclipse.core.resources.ResourcesPlugin;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IExtensionPoint;
import org.eclipse.core.runtime.IExtensionRegistry;
import org.eclipse.core.runtime.ILog;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.MultiStatus;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.Plugin;
import org.eclipse.core.runtime.Status;

import org.eclipse.core.runtime.dynamichelpers.ExtensionTracker;
import org.eclipse.core.runtime.dynamichelpers.IExtensionTracker;

import org.osgi.framework.Bundle;

/**
 * Plug-in adapter, used as plug-in abstraction.
 * 
 * @param <T>
 *            Plug-in type
 * 
 * @author Osama El Hosami
 */
@Contract
public final class PluginAdapter<T extends Plugin> implements Closeable {
	@Invar("null != $target")
	private final T _plugin;
	@Invar("null != $target")
	private final AbstractLogManager _logManager;

	private volatile boolean _isOpen;
	private IExtensionTracker _extTracker;

	/**
	 * Bundle state enumeration
	 */
	public enum State {
		INSTALLED, RESOLVED, STARTING, ACTIVE, STOPPING, UNINSTALLED;

		/**
		 * Map's bundle state value to equal enumeration.
		 * 
		 * @param value
		 *            Bundle state
		 * 
		 * @return Enumeration of bundle state
		 */
		static public State valueOf(int value) {
			State state = null;
			switch (value) {
			case Bundle.INSTALLED:
				state = INSTALLED;
				break;
			case Bundle.RESOLVED:
				state = RESOLVED;
				break;
			case Bundle.STARTING:
				state = STARTING;
				break;
			case Bundle.ACTIVE:
				state = ACTIVE;
				break;
			case Bundle.STOPPING:
				state = STOPPING;
				break;
			case Bundle.UNINSTALLED:
				break;
			default:
				throw new IllegalStateException();
			}
			return state;
		}
	};

	/**
	 * Creates a plug-in adapter
	 * 
	 * @param plugin
	 *            The adaptee
	 * @param logProperties
	 *            Configuration file of logger contained in plug-in bundle
	 */
	@Pre("null != plugin && null != logProperties")
	public PluginAdapter(final T plugin, final String logProperties) {
		this(plugin, new LogManager(plugin, logProperties));
	}

	/**
	 * Creates a plug-in adapter using given log manager.
	 * 
	 * @param plugin
	 *            The adaptee
	 * @param manager
	 *            Log manager
	 */
	@Pre("null != plugin && null != manager")
	private PluginAdapter(final T plugin, final AbstractLogManager manager) {
		_plugin = plugin;
		_logManager = manager;
		_isOpen = false;
	}

	/**
	 * Creates a plug-in adapter inheriting properties of given plug-in adapter.
	 * 
	 * @param plugin
	 *            The adaptee
	 * @param another
	 *            Another plug-in
	 */
	public PluginAdapter(final T plugin,
			final PluginAdapter<? extends Plugin> another) {
		this(plugin, new LogManagerProxy(plugin, another.getLogManager()));
	}

	/**
	 * Tests if plug-in adapter has been closed
	 * 
	 * @return True if satisfied otherwise false
	 */
	public boolean isClosed() {
		return !_isOpen;
	}

	/**
	 * Retrieve the plug-in adapter bundle state
	 * 
	 * @return The plug-in adapter bundle state
	 */
	public State getState() {
		return State.valueOf(getAdaptee().getBundle().getState());
	}

	/**
	 * Tests if plug-in adapter is in starting or active state
	 * 
	 * @return True if satisfied otherwise false
	 */
	public boolean isStartingOrActive() {
		return State.STARTING == getState() || State.ACTIVE == getState();
	}

	/**
	 * Tests if plug-in adapter is in stopping state
	 * 
	 * @return True if satisfied otherwise false
	 */
	public boolean isStopping() {
		return State.STOPPING == getState();
	}

	/**
	 * Initializes the plug-in adapter.
	 */
	@Pre("$this.isStartingOrActive() || !$this.isClosed()")
	@Post("!$this.isClosed()")
	public void init() {
		// double checked locking - works in JVM's greater or equal to version
		// 1.5
		if (isClosed()) {
			synchronized (this) {
				if (isClosed()) {
					_extTracker = new ExtensionTracker(getExtensionRegistry());
					_isOpen = _logManager.init();
				}
			}
		}
	}

	@Override
	@Pre("$this.isStopping()")
	@Post("$this.isClosed()")
	public void close() throws IOException {
		synchronized (this) {
			if (!isClosed()) {
				_logManager.close();
				_extTracker.close();
				_extTracker = null;
				_isOpen = false;
			}
		}
	}

	@Override
	public String toString() {
		return getPluginId();
	}

	/**
	 * Retrieves the wrapped plug-in.
	 * 
	 * @return The wrapped plug-in
	 */
	public T getAdaptee() {
		return _plugin;
	}

	/**
	 * Retrieves the plug-in identifier
	 * 
	 * @return The plug-in identifier
	 */
	@Post("null != $return")
	public String getPluginId() {
		return getAdaptee().getBundle().getSymbolicName();
	}

	/**
	 * Retrieves the plug-in log
	 * 
	 * @return The plug-in log
	 */
	@Post("null != $return")
	public ILog getLog() {
		return getAdaptee().getLog();
	}

	/**
	 * Retrieve workspace associated.
	 * 
	 * @return Workspace
	 */
	@Post("null != $return")
	public IWorkspace getWorkspace() {
		return ResourcesPlugin.getWorkspace();
	}

	/**
	 * Retrieves the plug-in log manager
	 * 
	 * @return The plug-in log manager
	 */
	@Post("null != $return")
	private AbstractLogManager getLogManager() {
		init();
		return _logManager;
	}

	/**
	 * Returns a new logger instance named as the first parameter using the
	 * default factory. If a logger of that name already exists, then it will be
	 * returned. Otherwise, a new logger will be instantiated and then linked
	 * with its existing ancestors as well as children.
	 * 
	 * @param name
	 *            Logger name
	 * 
	 * @return Logger
	 */
	@Pre("null != name")
	@Post("null != $return")
	public Logger getLogger(final String name) {
		return getLogManager().getLogger(name);
	}

	/**
	 * Retrieves the extension tracker instance.
	 * 
	 * @return The extension tracker instance
	 */
	@Post("null != $return")
	public IExtensionTracker getExtensionTracker() {
		init();
		return _extTracker;
	}

	/**
	 * Retrieves the extension registry instance.
	 * 
	 * @return The extension registry instance
	 */
	@Post("null != $return")
	public IExtensionRegistry getExtensionRegistry() {
		return Platform.getExtensionRegistry();
	}

	/**
	 * Retrieves the extension point of plug-in adapter.
	 * 
	 * @param extensionPointName
	 *            Name of extension point
	 * 
	 * @return Extension point or null
	 */
	@Pre("null != extensionPointName")
	public IExtensionPoint getExtensionPoint(final String extensionPointName) {
		return getExtensionRegistry().getExtensionPoint(getPluginId(),
				extensionPointName);
	}

	/**
	 * Panics - aborts any further execution, through throwing an unchecked
	 * exception.
	 * 
	 * @param message
	 *            Message or null
	 * @param e
	 *            Exception or null
	 */
	@Pre("null != message || null != e")
	public void panic(final String message, final Throwable e) {
		final IStatus status = createError(message, e);
		getLog().log(status);
		throw new RuntimeException(status.getMessage());
	}

	/**
	 * Creates a status representing a warning message.
	 * 
	 * @param message
	 *            Message or null
	 * @param e
	 *            Exception or null
	 * 
	 * @return Status
	 */
	@Pre("null != message || null != e")
	public IStatus createWarning(final String message, final Throwable e) {
		return new Status(IStatus.WARNING, getPluginId(), message, e);
	}

	/**
	 * Creates log entry representing a warning message.
	 * 
	 * @param message
	 *            Message
	 */
	@Pre("null != message")
	public void logWarning(final String message) {
		getLog().log(createWarning(message, null));
	}

	/**
	 * Reports a status representing a warning message.
	 * 
	 * @param message
	 *            Message
	 * @param e
	 *            Exception
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>Always</li>
	 *             </ul>
	 */
	@Pre("null != message || null != e")
	public void reportWarning(final String message, final Throwable e)
			throws CoreException {
		final IStatus status = createWarning(message, e);
		throw new CoreException(status);
	}

	/**
	 * Creates a status representing an error message.
	 * 
	 * @param message
	 *            Message or null
	 * @param e
	 *            Exception or null
	 * 
	 * @return Status
	 */
	@Pre("null != message || null != e")
	public IStatus createError(final String message, final Throwable e) {
		return new Status(IStatus.ERROR, getPluginId(), message, e);
	}

	/**
	 * Creates a status combined from multiple statuses.
	 * 
	 * @param statuses
	 *            Statuses
	 * @param message
	 *            Message
	 * 
	 * @return Status
	 */
	@Pre("null != statuses && null != message")
	public IStatus createError(final IStatus[] statuses, final String message) {
		return new MultiStatus(getPluginId(), 0, statuses, message, null);
	}

	/**
	 * Reports a status combined from multiple statuses.
	 * 
	 * @param statuses
	 *            Statuses
	 * @param message
	 *            Message
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>Always</li>
	 *             </ul>
	 */
	@Pre("null != statuses && null != message")
	public void reportError(final IStatus[] statuses, final String message)
			throws CoreException {
		final IStatus status = createError(statuses, message);
		final CoreException exception = new CoreException(status);
		throw exception;
	}

	/**
	 * Creates a status combined from multiple exceptions.
	 * 
	 * @param exceptions
	 *            Exceptions
	 * @param message
	 *            Message
	 * 
	 * @return Status
	 */
	@Pre("null != exceptions && null != message")
	public <E extends CoreException> IStatus createError(final E[] exceptions,
			final String message) {

		final IStatus[] status = new IStatus[exceptions.length];

		for (int i = 0; i < exceptions.length; i++) {
			status[i] = exceptions[i].getStatus();
		}

		return createError(status, message);
	}

	/**
	 * Reports a status combined from multiple exceptions.
	 * 
	 * @param exceptions
	 *            Exceptions
	 * @param message
	 *            Message
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>Always</li>
	 *             </ul>
	 */
	@Pre("null != exceptions && null != message")
	public void reportError(final CoreException[] exceptions,
			final String message) throws CoreException {
		final IStatus status = createError(exceptions, message);
		final CoreException exception = new CoreException(status);
		throw exception;
	}

	/**
	 * Creates a log entry representing an error message.
	 * 
	 * @param message
	 *            Message
	 */
	@Pre("null != message")
	public void logError(final String message) {
		getLog().log(createError(message, null));
	}

	/**
	 * Reports a status representing an error message.
	 * 
	 * @param message
	 *            Message or null
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>Always</li>
	 *             </ul>
	 */
	@Pre("null != message")
	public void reportError(final String message) throws CoreException {
		reportError(message, null);
	}

	/**
	 * Reports a status representing an error message.
	 * 
	 * @param message
	 *            Message or null
	 * @param e
	 *            Exception or null
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>Always</li>
	 *             </ul>
	 */
	@Pre("null != message || null != e")
	public void reportError(final String message, final Throwable e)
			throws CoreException {
		final IStatus status = createError(message, e);
		if (null != e) {
			e.printStackTrace();
		}
		throw new CoreException(status);
	}

	/**
	 * Creates a status representing an info message.
	 * 
	 * @param message
	 *            Message or null
	 * @param e
	 *            Exception or null
	 * 
	 * @return Status
	 */
	@Pre("null != message || null != e")
	public IStatus createInfo(final String message, final Throwable e) {
		return new Status(IStatus.INFO, getPluginId(), message, e);
	}

	/**
	 * Creates a log entry representing an info message.
	 * 
	 * @param message
	 *            Message
	 */
	@Pre("null != message")
	public void logInfo(final String message) {
		getLog().log(createInfo(message, null));
	}

	/**
	 * Reports a status representing an info message.
	 * 
	 * @param message
	 *            Message or null
	 * @param e
	 *            Exception or null
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>Always</li>
	 *             </ul>
	 */
	@Pre("null != message || null != e")
	public void reportInfo(final String message, final Throwable e)
			throws CoreException {
		final IStatus status = createInfo(message, e);
		throw new CoreException(status);
	}
}
