package de.fernunihagen.rtt.internal;

import java.io.Closeable;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;

import java.net.URLConnection;
import java.nio.channels.FileChannel;

import org.apache.log4j.Logger;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceVisitor;
import org.eclipse.core.resources.IWorkspaceRunnable;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubMonitor;

import de.fernunihagen.rtt.internal.util.AbstractFileNode;
import de.fernunihagen.rtt.internal.util.DirectoryNode;
import de.fernunihagen.rtt.internal.util.FileNode;
import de.fernunihagen.rtt.internal.util.FileNodeVisitor;
import de.fernunihagen.rtt.internal.util.MutableBoolean;

import static de.fernunihagen.rtt.internal.RTTCorePlugin.getRTTCorePlugin;
import static de.fernunihagen.rtt.internal.util.ClassUtil.getClassUtil;

/**
 * Non-UI resource helper.
 * 
 * @author Osama El Hosami
 */
@Contract
public final class ResourceUtil {

	private final static class ResourceUtilHolder {
		private static final ResourceUtil _DEFAULT = new ResourceUtil();
	}

	private static final class LoggerHolder {
		private static final Logger _DEFAULT = getRTTCorePlugin().getLogger(
				ResourceUtil.class.getName());
	}

	@Post("null != $return")
	private Logger getLogger() {
		return LoggerHolder._DEFAULT;
	}

	private ResourceUtil() {

	}

	/**
	 * Returns the shared instance
	 * 
	 * @return the shared instance
	 */
	public static ResourceUtil getResourceUtil() {
		return ResourceUtilHolder._DEFAULT;
	}

	/**
	 * Closes a closeable resource
	 * 
	 * @param closeable
	 *            Resource
	 * 
	 * @return True if closed otherwise failed
	 */
	public boolean close(final Closeable closeable) {
		boolean fClosed = false;
		if (closeable != null) {
			try {
				closeable.close();
				fClosed = true;
			} catch (IOException ioe) {
				getLogger().trace(ioe);
			}
		}
		return fClosed;
	}

	/**
	 * Determines the MIME type of a file
	 * 
	 * @param file
	 *            File
	 * 
	 * @return The MIME type
	 */
	@Pre("null != file && file.isFile()")
	@Post("null != $return")
	public String getMIMEType(final File file) {
		// TODO change implementation, there are faster ways to do this
		// take a look at http://www.rgagnon.com/javadetails/java-0487.html
		String MIMEType = "text/plain";
		final URLConnection connection;
		try {
			connection = file.toURI().toURL().openConnection();
			MIMEType = connection.getContentType();
		} catch (IOException e) {
			getLogger().trace(e);
		}
		return MIMEType;
	}

	/**
	 * Retrieves the file extension of a file
	 * 
	 * @param file
	 *            File
	 * 
	 * @return The file extension
	 */
	@Pre("null != file && file.isFile()")
	@Post("null != $return")
	public String getFileExtension(final File file) {
		String ext = "";
		if (!file.isDirectory()) {
			final String name = file.getName();
			final int pos = name.lastIndexOf(".");
			if (-1 != pos) {
				ext = name.substring(pos);
			}
		}
		return ext;
	}

	/**
	 * Test if file is a zip archive
	 * 
	 * @param file
	 *            File
	 * 
	 * @return True if satisfied otherwise false
	 */
	public boolean isZipOrJarArchive(final File file) {
		boolean fResult = null != file && file.isFile();
		if (fResult) {
			final String MIMEtype = getMIMEType(file);

			fResult = 0 == MIMEtype.compareTo("application/zip")
					|| 0 == MIMEtype.compareTo("application/jar");
		}

		return fResult;
	}

	/**
	 * Creates a folder.
	 * 
	 * @param destination
	 *            Location of folder
	 * @param fCreateParents
	 *            Flag indicating if parent folder creation is necessary
	 * 
	 * @return True in case folder created otherwise false
	 */
	@Pre("null != destination")
	public boolean createFolder(final IPath destination,
			final boolean fCreateParents) {
		boolean created;

		if (!fCreateParents) {
			created = new File(destination.toOSString()).mkdir();
		} else {
			created = new File(destination.toOSString()).mkdirs();
		}

		return created;
	}

	/**
	 * Deletes files and folders in the given file system location.
	 * 
	 * @param file
	 *            File
	 * @param fDeleteChildren
	 *            Flag indicating deletion of children is requested
	 * 
	 * @return True if deleted otherwise false
	 */
	@Pre("null != location")
	public boolean delete(final File file, final boolean fDeleteChildren) {
		final MutableBoolean status = new MutableBoolean(true);
		final FileNodeVisitor visitor = new FileNodeVisitor() {
			@Override
			public void visit(final DirectoryNode node) {
				if (fDeleteChildren) {
					super.visit(node);
				}
				getLogger().debug("Deleting " + node);
				status.setValue(status.getValue() & node.toFile().delete());
			}

			@Override
			public void visit(final FileNode node) {
				getLogger().debug("Deleting " + node);
				status.setValue(status.getValue() & node.toFile().delete());
			}
		};

		AbstractFileNode node = AbstractFileNode.create(file);
		node.accept(visitor);

		return status.getValue();
	}

	/**
	 * Deletes files and folders in the given file system location.
	 * 
	 * @param location
	 *            Location
	 * @param fDeleteChildren
	 *            Flag indicating deletion of children is requested
	 * 
	 * @return True if deleted otherwise false
	 */
	@Pre("null != location")
	public boolean delete(IPath location, final boolean fDeleteChildren) {
		return delete(location.toFile(), fDeleteChildren);
	}

	private <T> T as(final IAdaptable adaptable, final Class<? extends T> clazz) {
		return getClassUtil().as(adaptable, clazz);
	}

	/**
	 * Copies a resource with all its members to a file system location
	 * 
	 * @param resource
	 *            Resource to be copied
	 * @param destination
	 *            Destination
	 * @param pm
	 *            The progress monitor to use for reporting progress to the
	 *            user. Accepts <code>null</code>, indicating that no progress
	 *            should be reported and that the operation cannot be canceled
	 * 
	 * @return Status code
	 */
	public IStatus copy(final IResource resource, final IPath destination,
			final IProgressMonitor pm) {
		return copy(resource, destination, IResource.DEPTH_INFINITE, 0, pm);
	}

	private void reportError(final Throwable e) throws CoreException {
		getRTTCorePlugin().reportError(null, e);
	}

	/**
	 * Copies a resource to a file system location.
	 * 
	 * @param resource
	 *            Resource to be copied
	 * @param destination
	 *            Destination
	 * @param depth
	 *            The depth to which members of this resource should be visited.
	 *            One of {@link IResource#DEPTH_ZERO},
	 *            {@link IResource#DEPTH_ONE}, or
	 *            {@link IResource#DEPTH_INFINITE}.
	 * @param memberFlags
	 *            Bit-wise or of member flag constants (
	 *            {@link IContainer#INCLUDE_PHANTOMS},
	 *            {@link IContainer#INCLUDE_TEAM_PRIVATE_MEMBERS},
	 *            {@link IContainer#INCLUDE_HIDDEN} and
	 *            {@link IContainer#EXCLUDE_DERIVED}) indicating which members
	 *            are of interest
	 * @param pm
	 *            The progress monitor to use for reporting progress to the
	 *            user. Accepts <code>null</code>, indicating that no progress
	 *            should be reported and that the operation cannot be canceled
	 * 
	 * @return Status code
	 */
	@Pre("null != resource && null != destination")
	@Post("null != $return")
	public IStatus copy(final IResource resource, final IPath destination,
			final int depth, final int memberFlags, final IProgressMonitor pm) {

		IStatus status = Status.OK_STATUS;

		try {
			final SubMonitor monitor = SubMonitor.convert(pm,
					"Copying resource(s)", IProgressMonitor.UNKNOWN);
			final IResourceVisitor visitor = new IResourceVisitor() {
				@Invar("null != $target")
				public IPath curDestination = destination;

				@Pre("null != resource")
				private boolean copyContainer(final IContainer resource) {
					curDestination = curDestination.append(resource.getName());
					final String message = "Creating folder "
							+ curDestination.toPortableString();
					monitor.subTask(message);
					getLogger().debug(message);
					createFolder(curDestination, false);
					monitor.worked(1);
					return true;
				}

				@Pre("null != resource")
				private boolean copyFile(final IFile resource)
						throws CoreException {
					final File source = new File(resource.getLocationURI());
					final String message = "Copying file "
							+ resource.getLocation().toPortableString();
					monitor.subTask(message);
					getLogger().debug(message);
					copy(source, curDestination.toFile());
					monitor.worked(1);
					return false;
				}

				@Override
				public boolean visit(final IResource resource)
						throws CoreException {
					final boolean fFile = IResource.FILE == resource.getType();
					final boolean fVisitMember = (fFile) ? copyFile(as(
							resource, IFile.class)) : copyContainer(as(
							resource, IContainer.class));

					if (monitor.isCanceled()) {
						throw new OperationCanceledException();
					}

					return fVisitMember;
				}
			};

			final IWorkspaceRunnable operation = new IWorkspaceRunnable() {
				@Override
				public void run(IProgressMonitor pm) throws CoreException {
					resource.accept(visitor, depth, memberFlags);
				}
			};

			resource.getWorkspace().run(operation, null);

		} catch (CoreException cx) {
			getLogger().trace(cx);
			delete(destination, true);
			status = cx.getStatus();
		} catch (OperationCanceledException oce) {
			getLogger().trace(oce);
			delete(destination, true);
			status = Status.CANCEL_STATUS;
		} finally {
			if (null != pm) {
				pm.done();
			}
		}

		return status;
	}

	/**
	 * Copies a file from source to destination location.
	 * 
	 * @param source
	 *            source location
	 * @param destination
	 *            destination location
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If an error occurs during operation</li>
	 *             </ul>
	 */
	public void copy(final File source, final File destination)
			throws CoreException {
		FileInputStream is = null;
		FileOutputStream os = null;
		FileChannel inChannel = null;
		FileChannel outChannel = null;

		try {
			is = new FileInputStream(source);
			os = new FileOutputStream(destination);
			inChannel = is.getChannel();
			outChannel = os.getChannel();
			inChannel.transferTo(0, inChannel.size(), outChannel);
		} catch (IOException e) {
			reportError(e);
		} finally {
			close(inChannel);
			close(outChannel);
			close(is);
			close(os);
		}
	}
}
