package de.fernunihagen.rtt.internal.jdt;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.apache.log4j.Logger;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.resources.IProject;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.SubMonitor;

import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IPackageFragment;
import org.eclipse.jdt.core.JavaCore;

import de.fernunihagen.rtt.framework.ElementFilterOp;
import de.fernunihagen.rtt.framework.IElementCollector;
import de.fernunihagen.rtt.framework.IElementFilter;

import static de.fernunihagen.rtt.internal.RTTCorePlugin.getRTTCorePlugin;

/**
 * Compilation unit collector.
 * 
 * @author Osama El Hosami
 */
@Contract
public class CompilationUnitCollector implements
		IElementCollector<ICompilationUnit> {
	@Invar("null != $target")
	private final List<ICompilationUnit> _list;
	@Invar("null != $target")
	private final IElementFilter<IAdaptable> _filter;
	@Invar("null != $target")
	private final IProject[] _scope;

	private static final class LoggerHolder {
		private static final Logger _DEFAULT = getRTTCorePlugin().getLogger(
				CompilationUnitCollector.class.getName());
	}

	@Post("null != $return")
	private Logger getLogger() {
		return LoggerHolder._DEFAULT;
	}

	/**
	 * Create a compilation unit collector, used to fill given list with
	 * matching elements. List has to be synchronized in case collector is used
	 * concurrently.
	 * 
	 * @param list
	 *            List
	 * @param filter
	 *            Filter
	 * @param scope
	 *            Projects in search scope
	 */
	@Pre("null != list && null != filter && null != scope")
	public CompilationUnitCollector(final List<ICompilationUnit> list,
			final IElementFilter<IAdaptable> filter, final IProject[] scope) {
		_list = list;
		_filter = ElementFilterOp.and(filter, CompilationUnitFilter
				.getDefault());
		_scope = scope;
	}

	/**
	 * Creates a compilation unit collector, used to collect matching elements.
	 * 
	 * @param filter
	 *            Filter
	 * @param scope
	 *            Projects in search scope
	 */
	@Pre("null != filter && null != scope")
	public CompilationUnitCollector(final IElementFilter<IAdaptable> filter,
			final IProject[] scope) {
		this(new ArrayList<ICompilationUnit>(), filter, scope);
	}

	/**
	 * Creates a compilation unit collector.
	 * 
	 * @param scope
	 *            Projects in search scope
	 */
	@Pre("null != scope")
	public CompilationUnitCollector(final IProject[] scope) {
		this(IElementFilter.ALL, scope);
	}

	@Pre("null != units && null != monitor")
	private void collect(final ICompilationUnit[] units,
			final SubMonitor monitor) {

		monitor.beginTask("Accepting Java compilation unit(s)", units.length);
		getLogger().debug("Accepting Java compilation unit(s)");

		for (ICompilationUnit unit : units) {
			monitor.subTask(unit.getElementName());

			final boolean fAccept = _filter.accept(unit);

			if (getLogger().isDebugEnabled()) {
				final String message = String.format("[/U %s (%s)]", unit
						.getElementName(), fAccept);
				getLogger().debug(message);
			}

			if (fAccept) {
				_list.add(unit);
			}

			monitor.worked(1);
		}

		getLogger().debug("Accept done");
	}

	@Pre("null != fragments && null != monitor")
	private void collect(final IPackageFragment[] fragments,
			final SubMonitor monitor) throws CoreException {

		monitor.beginTask("Analyzing fragment(s)", 2 * fragments.length);
		getLogger().debug("Analyzing fragment(s)");

		for (IPackageFragment fragment : fragments) {
			boolean fIsArchive = fragment.isReadOnly();
			if (!fIsArchive) {
				monitor.subTask(fragment.getElementName());
				if (getLogger().isDebugEnabled()) {
					final String message = String.format("[F/ %s]", fragment
							.getElementName());
					getLogger().debug(message);
				}

				fragment.open(monitor.newChild(1));
				collect(fragment.getCompilationUnits(), monitor.newChild(1));
			}
		}

		getLogger().debug("Analysis done");
	}

	@Override
	public void collect(final IProgressMonitor pm) throws CoreException {
		try {
			clear();

			final IProject[] projects = getScope();
			final SubMonitor monitor = SubMonitor.convert(pm,
					"Collecting Java compilation unit(s)", 2 * projects.length);

			getLogger().debug("Collecting Java compilation unit(s)");

			for (IProject project : projects) {
				monitor.subTask(project.getName());
				project.open(monitor.newChild(1));

				final boolean fIsJavaProject = project
						.hasNature(JavaCore.NATURE_ID);

				if (getLogger().isDebugEnabled()) {
					final String message = String.format("[/P %s (%s)]",
							project.getName(), fIsJavaProject);
					getLogger().debug(message);
				}

				if (fIsJavaProject) {
					final IJavaProject javaProject = JavaCore.create(project);

					collect(javaProject.getPackageFragments(), monitor
							.newChild(1));
				} else {
					monitor.worked(1);
				}
			}

			if (getLogger().isDebugEnabled()) {
				final String message = String.format("Collection done (%d)",
						_list.size());
				getLogger().debug(message);
			}

		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}

	@Override
	public List<? extends ICompilationUnit> getMatches() {
		return Collections.unmodifiableList(_list);
	}

	@Override
	public IProject[] getScope() {
		return _scope;
	}

	@Override
	public void clear() {
		_list.clear();
	}

}
