package de.fernunihagen.rtt.internal.jdt;

import static de.fernunihagen.rtt.internal.util.ClassUtil.getClassUtil;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.runtime.IAdaptable;

import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.JavaCore;

import de.fernunihagen.rtt.framework.IElementFilter;

/**
 * Excluded java element filter.
 * 
 * @author Osama El Hosami
 */
@Contract
public final class ExcludedJavaElementFilter implements
		IElementFilter<IAdaptable> {

	@Invar("null != $target")
	private final Set<IJavaElement> _excluded;

	/**
	 * Retrieves the corresponding list of java elements for given identifier
	 * list.
	 * 
	 * @param elementHandleIdentifiers
	 *            Element handle identifier
	 * 
	 * @return List of java elements
	 */
	public static IJavaElement[] getElementsByHandle(
			String[] elementHandleIdentifiers) {
		final Set<IJavaElement> elements = new HashSet<IJavaElement>(
				elementHandleIdentifiers.length);
		for (String elementHandleIdentifier : elementHandleIdentifiers) {
			IJavaElement element = JavaCore.create(elementHandleIdentifier,
					null);
			if (null != element) {
				elements.add(element);
			}
		}

		return elements.toArray(new IJavaElement[elements.size()]);
	}

	/**
	 * Creates an excluded java element filter.
	 */
	public ExcludedJavaElementFilter() {
		_excluded = new HashSet<IJavaElement>();
	}

	/**
	 * Creates an excluded java element filter.
	 * 
	 * @param excluded
	 *            List of excluded java elements
	 */
	@Pre("null != excluded")
	public ExcludedJavaElementFilter(final IJavaElement[] excluded) {
		_excluded = new HashSet<IJavaElement>(excluded.length);
		addAll(excluded);
	}

	/**
	 * Creates an excluded java element filter.
	 * 
	 * @param elementHandleIdentifiers
	 *            List of excluded element handle identifiers
	 */
	@Pre("null != excluded")
	public ExcludedJavaElementFilter(final String[] elementHandleIdentifiers) {
		this(getElementsByHandle(elementHandleIdentifiers));
	}

	private <T> T as(final IAdaptable adaptable, final Class<? extends T> clazz) {
		return getClassUtil().as(adaptable, clazz);
	}

	/**
	 * Adds a list of element handle
	 * 
	 * @param elementHandleIdentifiers
	 *            List of excluded element handle identifiers
	 */
	public void addAll(final String[] elementHandleIdentifiers) {
		addAll(getElementsByHandle(elementHandleIdentifiers));
	}

	/**
	 * Adds a list of element handle
	 * 
	 * @param excluded
	 *            List of excluded java elements
	 */
	public void addAll(final IJavaElement[] excluded) {
		_excluded.addAll(Arrays.asList(excluded));
	}

	/**
	 * Adds an element to filter configuration.
	 * 
	 * @param element
	 *            Element
	 */
	@Pre("null != element")
	public void add(final IJavaElement element) {
		_excluded.add(element);
	}

	/**
	 * Removes an element from filter configuration.
	 * 
	 * @param element
	 *            Element
	 */
	@Pre("null != element")
	public void remove(final IJavaElement element) {
		_excluded.remove(element);
	}

	/**
	 * Clears all elements from filter configuration.
	 */
	public void clear() {
		_excluded.clear();
	}

	/**
	 * Retrieves the list of excluded element handle identifiers.
	 * 
	 * @return List of element handle identifiers
	 */
	@Post("null != $return")
	public String[] getHandleIdentifiers() {
		final List<String> identifiers = new ArrayList<String>(_excluded.size());
		for (IJavaElement element : _excluded) {
			identifiers.add(element.getHandleIdentifier());
		}
		return identifiers.toArray(new String[identifiers.size()]);
	}

	/**
	 * Retrieves the list of excluded element handle identifiers contained in
	 * given project.
	 * 
	 * @param project
	 *            Project
	 * 
	 * @return List of element handle identifiers
	 */
	@Pre("null != project")
	@Post("null != $return")
	public String[] getHandleIdentifiers(final IJavaProject project) {
		final List<String> identifiers = new ArrayList<String>(_excluded.size());
		for (IJavaElement element : _excluded) {
			if (element.getJavaProject().equals(project)) {
				identifiers.add(element.getHandleIdentifier());
			}
		}
		return identifiers.toArray(new String[identifiers.size()]);
	}

	@Override
	public boolean accept(final IAdaptable element) {
		final IJavaElement javaElement = as(element, IJavaElement.class);
		return !_excluded.contains(javaElement);
	}
}