package de.fernunihagen.rtt.internal.jdt;

import static de.fernunihagen.rtt.internal.RTTCorePlugin.getRTTCorePlugin;

import org.apache.log4j.Logger;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.SubMonitor;

import org.eclipse.debug.core.DebugPlugin;
import org.eclipse.debug.core.ILaunchConfiguration;
import org.eclipse.debug.core.ILaunchConfigurationWorkingCopy;
import org.eclipse.debug.core.ILaunchManager;

import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.junit.JUnitCore;
import org.eclipse.jdt.junit.TestRunListener;

import org.eclipse.jdt.junit.model.ITestCaseElement;
import org.eclipse.jdt.junit.model.ITestRunSession;
import org.eclipse.jdt.junit.model.ITestElement.Result;

import de.fernunihagen.rtt.internal.jdt.JUnitTestRunNotification.Type;
import de.fernunihagen.rtt.internal.util.Subject;

/**
 * Abstract JUnit test run.
 * 
 * @param <S>
 *            Test run subject type
 * 
 * @author Osama El Hosami
 */
@Contract
public abstract class JUnitTestRun<S>
		extends
		Subject<JUnitTestRun<S>, IJUnitTestRunListener<S>, JUnitTestRunNotification> {

	@Invar("null != $target")
	private final ILaunchConfiguration _launchConfiguration;
	@Invar("null != $target")
	private State _state;
	@Invar("null != $target")
	private final Listener _listener;
	@Invar("null != $target")
	private Result _result;
	private S _subject;

	private enum State {
		INIT, READY, ACTIVE, FINISHED
	}

	private boolean contains(final ITestCaseElement element) {
		return contains(element.getTestRunSession());
	}

	private boolean contains(final ITestRunSession session) {
		return session.getTestRunName().equals(_launchConfiguration.getName());
	}

	private final class Listener extends TestRunListener {
		@Override
		public void sessionStarted(final ITestRunSession session) {
			// filter notification of other launch configurations
			if (!contains(session)) {
				return;
			}

			_result = Result.UNDEFINED;
			setChanged();
			notifyObservers(new JUnitTestRunNotification(Type.SESSION_STARTED,
					session));
		}

		@Override
		public void sessionFinished(final ITestRunSession session) {
			if (!contains(session)) {
				return;
			}

			setChanged();
			notifyObservers(new JUnitTestRunNotification(Type.SESSION_FINISHED,
					session));
			_result = session.getTestResult(true);
			notifyWaiting();
		}

		@Override
		public void testCaseStarted(final ITestCaseElement testCaseElement) {
			if (!contains(testCaseElement)) {
				return;
			}

			setChanged();
			notifyObservers(new JUnitTestRunNotification(Type.TESTCASE_STARTED,
					testCaseElement));
		}

		@Override
		public void testCaseFinished(final ITestCaseElement testCaseElement) {
			if (!contains(testCaseElement)) {
				return;
			}

			setChanged();
			notifyObservers(new JUnitTestRunNotification(
					Type.TESTCASE_FINISHED, testCaseElement));
		}
	}

	private static final class LoggerHolder {
		private static final Logger _DEFAULT = getRTTCorePlugin().getLogger(
				JUnitTestRun.class.getName());
	}

	@Post("null != $return")
	private Logger getLogger() {
		return LoggerHolder._DEFAULT;
	}

	/**
	 * Retrieves the shared launch manager instance.
	 * 
	 * @return Launch manager
	 */
	protected static ILaunchManager getLaunchManager() {
		return DebugPlugin.getDefault().getLaunchManager();
	}

	/**
	 * Creates a JUnit test run.
	 * 
	 * @param wc
	 *            launch configuration working copy
	 * @param subject
	 *            Subject
	 */
	@Pre("null != wc")
	public JUnitTestRun(final ILaunchConfigurationWorkingCopy wc,
			final S subject) throws CoreException {
		this(wc.doSave(), subject);
	}

	/**
	 * Creates a JUnit test run.
	 * 
	 * @param launchConfiguration
	 *            launch configuration
	 * @param subject
	 *            Subject
	 */
	@Pre("null != launchConfiguration")
	public JUnitTestRun(final ILaunchConfiguration launchConfiguration,
			final S subject) {
		_launchConfiguration = launchConfiguration;
		_state = State.INIT;
		_listener = new Listener();
		_result = Result.UNDEFINED;
		_subject = subject;
	}

	private void notifyWaiting() {
		synchronized (this) {
			_state = State.FINISHED;
			notifyAll();
		}
	}

	private boolean initImpl() throws CoreException {
		if (State.INIT == _state) {
			init();
		}
		if (State.INIT == _state || State.FINISHED == _state) {
			_state = State.READY;
		}

		return State.READY == _state;
	}

	/**
	 * Initializes a test run.
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If an error occurs</li>
	 *             </ul>
	 */
	protected void init() throws CoreException {
		JUnitCore.addTestRunListener(_listener);
	}

	/**
	 * Starts the test run.
	 * 
	 * @param pm
	 *            The progress monitor to use for reporting progress to the
	 *            user. It is the caller's responsibility to call done() on the
	 *            given monitor. Accepts <code>null</code>, indicating that no
	 *            progress should be reported and that the operation cannot be
	 *            canceled
	 * 
	 * @return true if started otherwise false
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If an error occurs</li>
	 *             </ul>
	 */
	public final boolean start(final IProgressMonitor pm) throws CoreException {
		boolean fStart = false;
		try {
			final SubMonitor monitor = SubMonitor.convert(pm);

			synchronized (this) {
				fStart = initImpl();
				if (fStart) {
					_launchConfiguration.launch(ILaunchManager.RUN_MODE,
							monitor.newChild(1), false, true);

					if (monitor.isCanceled()) {
						throw new OperationCanceledException();
					}

					_state = State.ACTIVE;
				}
			}
		} finally {
			if (null != pm) {
				pm.done();
			}
		}
		return fStart;
	}

	/**
	 * Disposes all internal resources.
	 * 
	 * @param fDelete
	 *            Flag indicating deletion of configuration requested
	 */
	public void dispose(boolean fDelete) {
		try {
			synchronized (this) {
				JUnitCore.removeTestRunListener(_listener);

				if (fDelete) {
					_launchConfiguration.delete();
				}

				_state = State.INIT;
			}
		} catch (CoreException e) {
			getLogger().trace(
					"Failed to delete launch configuration '"
							+ _launchConfiguration.getName() + "'", e);
		}
	}

	/**
	 * Waits until test run is finished.
	 */
	public void join() {
		synchronized (this) {
			while (State.ACTIVE == _state) {
				try {
					wait();
				} catch (InterruptedException e) {
				}
			}
		}
	}

	@Override
	@Post("null != $return")
	public String toString() {
		return _launchConfiguration.getName();
	}

	@Override
	@Post("null != $return")
	protected JUnitTestRun<S> getThis() {
		return this;
	}

	/**
	 * Retrieves the overall test result
	 * 
	 * @return Test result
	 */
	@Post("null != $return")
	public Result getTestResult() {
		join();
		return _result;
	}

	/**
	 * Retrieves the test run subject.
	 * 
	 * @return Subject or null if not applicable
	 */
	public S getSubject() {
		return _subject;
	}

	/**
	 * Retrieves the launched project
	 * 
	 * @return Project
	 */
	public abstract IJavaProject getProject();
}
