package de.fernunihagen.rtt.internal.jdt.oracle;

import static de.fernunihagen.rtt.internal.RTTCorePlugin.getRTTCorePlugin;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.runtime.IAdaptable;

import org.eclipse.jdt.junit.model.ITestCaseElement;
import org.eclipse.jdt.junit.model.ITestElement;
import org.eclipse.jdt.junit.model.ITestRunSession;
import org.eclipse.jdt.junit.model.ITestSuiteElement;

import de.fernunihagen.rtt.framework.IElementCollector;
import de.fernunihagen.rtt.framework.ITestActor;
import de.fernunihagen.rtt.framework.ITestElementContainer;

import de.fernunihagen.rtt.framework.ITestElement.ProgressState;

import de.fernunihagen.rtt.framework.oracle.OracleRunDescription;
import de.fernunihagen.rtt.framework.oracle.OracleRunState;

import de.fernunihagen.rtt.internal.jdt.JUnitTestRun;
import de.fernunihagen.rtt.internal.util.ReversedList;

/**
 * JUnit oracle test run context.
 * 
 * @author Osama El Hosami
 */
@Contract
public final class JUnitOracleRunState extends OracleRunState {

	@Invar("null != $target")
	private final List<JUnitTestRun<JUnitOracleRunState>> _testruns;
	@Invar("null != $target")
	private final Map<String, ITestElementContainer> _activeContainers;
	@Invar("null != $target")
	private final Map<String, JUnitTestCaseElement> _activeCases;

	private static final class LoggerHolder {
		private static final Logger _DEFAULT = getRTTCorePlugin().getLogger(
				JUnitOracleRunState.class.getName());
	}

	@Post("null != $return")
	private Logger getLogger() {
		return LoggerHolder._DEFAULT;
	}

	/**
	 * Creates a JUnit oracle test run context.
	 * 
	 * @param originator
	 *            Originator
	 * @param collector
	 *            Collector
	 * @param description
	 *            Test description
	 */
	@Pre("null != originator && null != collector && null != description")
	public JUnitOracleRunState(final ITestActor originator,
			final IElementCollector<? extends IAdaptable> collector,
			final OracleRunDescription description) {
		super(originator, collector, description);
		_testruns = new ArrayList<JUnitTestRun<JUnitOracleRunState>>();
		_activeContainers = new HashMap<String, ITestElementContainer>();
		_activeCases = new HashMap<String, JUnitTestCaseElement>();
	}

	@Override
	public void reset() {
		_activeCases.clear();
		_activeContainers.clear();

		final Iterable<JUnitTestRun<JUnitOracleRunState>> reversed = ReversedList
				.reversed(_testruns);

		for (JUnitTestRun<JUnitOracleRunState> testrun : reversed) {
			testrun.deleteObservers();
			testrun.dispose(true);
		}
	}

	/**
	 * Retrieves the list of test runs.
	 * 
	 * @return List of test runs
	 */
	@Post("null != $return")
	public List<JUnitTestRun<JUnitOracleRunState>> getTestRuns() {
		return _testruns;
	}

	/**
	 * Adds a test element container to list of active test element container.
	 * 
	 * @param container
	 *            Test element container
	 */
	@Pre("null != container")
	public void addContainer(final ITestElementContainer container) {
		_activeContainers.put(container.getLabel(), container);
	}

	/**
	 * Removes a test element container from list of active test element
	 * container.
	 * 
	 * @param label
	 *            Label of test element container
	 */
	@Pre("null != testelement")
	public void removeContainer(final String label) {
		final ITestElementContainer container = _activeContainers.remove(label);
		container.setProgressState(ProgressState.COMPLETED);

		if (getLogger().isDebugEnabled()) {
			final String message = String.format("%s (%s;%s;%.3fs)", container,
					container.getTestResult(),
					container.getProgressState(true),
					container.getElapsedTimeInSeconds());
			getLogger().debug(message);
		}
	}

	/**
	 * Removes an active test element container from list of active test element
	 * container.
	 * 
	 * @param testelement
	 *            JUnit Test element
	 */
	@Pre("null != testelement")
	public void removeContainer(final ITestSuiteElement testelement) {
		removeContainer(testelement.getSuiteTypeName());
	}

	/**
	 * Retrieves the parent test element container for given JUnit model
	 * element.
	 * 
	 * @param testelement
	 *            JUnit Test element
	 * 
	 * @return Test element container
	 */
	public ITestElementContainer getParentContainer(
			final ITestElement testelement) {
		if (null == testelement) {
			return null;
		}

		ITestElementContainer parent = getContainer(testelement
				.getParentContainer());
		if (null != parent) {
			return parent;
		}

		return getContainer(testelement.getTestRunSession());
	}

	/**
	 * Retrieves the test element container for given JUnit model element.
	 * 
	 * @param testelement
	 *            JUnit Test element
	 * 
	 * @return Test element container
	 */
	public ITestElementContainer getContainer(final ITestElement testelement) {
		if (null == testelement) {
			return null;
		}

		if (testelement instanceof ITestRunSession) {
			return _activeContainers.get(((ITestRunSession) testelement)
					.getTestRunName());
		}

		if (testelement instanceof ITestSuiteElement) {
			return _activeContainers.get(((ITestSuiteElement) testelement)
					.getSuiteTypeName());
		}

		return getContainer(testelement.getParentContainer());
	}

	/**
	 * Sets a test element to list of active test elements.
	 * 
	 * @param testcase
	 *            Test element
	 */
	@Pre("null != testcase")
	public void addElement(final JUnitTestCaseElement testcase) {
		_activeCases.put(testcase.getLabel(), testcase);
	}

	private String getKey(final ITestCaseElement testelement) {
		final StringBuffer buffer = new StringBuffer();
		buffer.append(testelement.getTestClassName());
		buffer.append(".");
		buffer.append(testelement.getTestMethodName());

		return buffer.toString();
	}

	/**
	 * Retrieves the test element for given JUnit test element.
	 * 
	 * @param testelement
	 *            JUnit Test element
	 * 
	 * @return Test element
	 */
	@Pre("null != testelement")
	public JUnitTestCaseElement getElement(final ITestCaseElement testelement) {
		return _activeCases.get(getKey(testelement));
	}

	/**
	 * Removes a test element from list of active test elements.
	 * 
	 * @param testelement
	 *            JUnit Test element
	 */
	@Pre("null != testelement")
	public void removeElement(final ITestCaseElement testelement) {
		final JUnitTestCaseElement element = _activeCases
				.remove(getKey(testelement));

		if (!element.getTestResult().isSuccess()) {
			getParentContainer(testelement).add(element);
		}

		element.setProgressState(ProgressState.COMPLETED);

		if (getLogger().isDebugEnabled()) {
			final String message = String.format("%s (%s;%s;%.3fs)", element,
					element.getTestResult(), element.getProgressState(true),
					element.getElapsedTimeInSeconds());
			getLogger().debug(message);
		}
	}
}
