package de.fernunihagen.rtt.internal.runtime;

import static de.fernunihagen.rtt.internal.RTTCorePlugin.getRTTCorePlugin;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.SubMonitor;

import de.fernunihagen.rtt.framework.ITestActor;
import de.fernunihagen.rtt.framework.ITestConfiguration;
import de.fernunihagen.rtt.framework.ITestContext;
import de.fernunihagen.rtt.framework.ITestDescription;
import de.fernunihagen.rtt.framework.ITestElementContainer;
import de.fernunihagen.rtt.framework.ITestRunSession;

import de.fernunihagen.rtt.framework.model.TestContext;
import de.fernunihagen.rtt.framework.oracle.IOracleRunner;

import de.fernunihagen.rtt.framework.testadapter.IRefactoringTestAdapter;

/**
 * Refactoring test runner run state.
 * 
 * @author Osama El Hosami
 */
@Contract
public final class RefactoringTestRunnerRunState extends
		TestRunnerRunState<ITestActor> {

	@Invar("0 <= $target")
	private int _current;
	@Invar("0 <= $target")
	private int _next;
	@Invar("null != $target")
	private final List<CoreException> _errors;
	private boolean _fIsStalled;
	@Invar("null != $target")
	private final IOracleRunner _oracleRunner;
	private ITestContext _undoContext;

	private static ITestConfiguration[] getConfiguration(
			final ITestActor actor, final ITestRunSession session) {
		return session.getConfiguration(actor).getChildren(
				ITestDescription.TYPE_ACTORS_TESTADAPTER);
	}

	/**
	 * Creates a refactoring test runner run state.
	 * 
	 * @param originator
	 *            Originator
	 * @param session
	 *            Test run session
	 * @param refactoringTestAdapters
	 *            Map of refactoring test adapters
	 * @param oracleRunner
	 *            Oracle runner
	 */
	@Pre("null != originator && null != session && null != refactoringTestAdapters && null != oracleRunner")
	public RefactoringTestRunnerRunState(final ITestActor originator,
			final ITestRunSession session,
			final Map<String, IRefactoringTestAdapter> refactoringTestAdapters,
			final IOracleRunner oracleRunner) {
		super(originator, refactoringTestAdapters, getConfiguration(originator,
				session));
		_next = _current = 0;
		_errors = new ArrayList<CoreException>();
		_fIsStalled = false;
		_oracleRunner = oracleRunner;
		_undoContext = null;
	}

	@Override
	public void reset() {
		_errors.clear();
		_next = _current = 0;
		_fIsStalled = false;
		_undoContext = null;
	}

	@Override
	public void initialize(final ITestElementContainer parent,
			final IProgressMonitor pm) throws CoreException {
		try {
			final SubMonitor monitor = SubMonitor.convert(pm);
			createOracleRunnerContext(parent, monitor.newChild(1));
			super.initialize(parent, monitor.newChild(1));
		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}

	@Override
	@Post
	protected IRefactoringTestAdapter[] getEmptyArray() {
		return new IRefactoringTestAdapter[0];
	}

	private void reportError(final String message, final Throwable exception)
			throws CoreException {
		getRTTCorePlugin().reportError(message, exception);
	}

	private void reportError(final CoreException[] exceptions,
			final String message) throws CoreException {
		getRTTCorePlugin().reportError(exceptions, message);
	}

	/**
	 * Sets current undo context
	 * 
	 * @param context
	 *            Test context
	 */
	public void setUndoContext(final ITestContext context) {
		_undoContext = context;
	}

	/**
	 * Retrieves current undo context.
	 * 
	 * @return Test context or null
	 */
	public ITestContext getUndoContext() {
		return _undoContext;
	}

	/**
	 * Creates a oracle runner test context.
	 * 
	 * @param parent
	 *            Parent test element container
	 * @param pm
	 *            The progress monitor to use for reporting progress to the
	 *            user. It is the caller's responsibility to call done() on the
	 *            given monitor. Accepts <code>null</code>, indicating that no
	 *            progress should be reported and that the operation cannot be
	 *            canceled
	 * 
	 * @return Test context
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If an error occurs</li>
	 *             </ul>
	 */
	@Pre("null != parent")
	@Post("null != $return")
	public final TestContext createOracleRunnerContext(
			final ITestElementContainer parent, final IProgressMonitor pm)
			throws CoreException {
		return createContext(_oracleRunner, parent, null);
	}

	/**
	 * Retrieves the remaining work.
	 * 
	 * @return Work remaining
	 */
	public int getWorkRemaining() {
		return _actors.size() - _next;
	}

	/**
	 * Tests if run has errors.
	 * 
	 * @return true if run generated errors
	 */
	public boolean hasError() {
		return !_errors.isEmpty();
	}

	/**
	 * Reports a status combined from multiple exceptions occurred during run.
	 * 
	 * @param message
	 *            Message
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>Always</li>
	 *             </ul>
	 */
	public void reportError(final String message) throws CoreException {
		final CoreException[] errors = _errors
				.toArray(new CoreException[_errors.size()]);
		_errors.clear();
		reportError(errors, message);
	}

	/**
	 * Adds an exception.
	 * 
	 * @param e
	 *            exception
	 */
	public void addError(final CoreException e) {
		_errors.add(e);
	}

	/**
	 * Retrieves the current test adapter.
	 * 
	 * @return Test adapter
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             If no such element exists
	 *             </ul>
	 */
	@Post("null != $return")
	public IRefactoringTestAdapter getTestAdapter() throws CoreException {
		if (_current < 0 || _current >= _actors.size()) {
			reportError(null, new NoSuchElementException());
		}
		return (IRefactoringTestAdapter) _actors.get(_current);
	}

	/**
	 * Retrieves the next test adapter.
	 * 
	 * @return Test adapter
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             If no such element exists
	 *             </ul>
	 */
	@Post("null != $return")
	public IRefactoringTestAdapter nextTestAdapter() throws CoreException {
		if (!hasNext()) {
			reportError(null, new NoSuchElementException());
		}
		_current = _next;
		return (IRefactoringTestAdapter) _actors.get(_current);
	}

	/**
	 * Test if run state has more test adapters.
	 * 
	 * @return true if run state has more test adapters
	 */
	public boolean hasNext() {
		return _next >= 0 && _next < _actors.size();
	}

	/**
	 * Tests stall execution state
	 * 
	 * @param fState
	 *            State value
	 */
	public void setStalled(boolean fState) {
		_fIsStalled = fState;
	}

	/**
	 * Tests whether execution is stalled or not.
	 * 
	 * @return true if execution is stalled
	 */
	public boolean isStalled() {
		return _fIsStalled;
	}

	/**
	 * Seeks to next test adapter if execution is not stalled.
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             If no such element exists
	 *             </ul>
	 */
	@Post("null != $return")
	public void seekNext() throws CoreException {
		if (!hasNext()) {
			reportError(null, new NoSuchElementException());
		}
		if (!isStalled()) {
			_next++;
		}
	}

	@Override
	public boolean isActive() {
		return hasNext();
	}
}
