package de.fernunihagen.rtt.internal.runtime;

import static de.fernunihagen.rtt.internal.RTTCorePlugin.getRTTCorePlugin;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Logger;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.runtime.CoreException;

import de.fernunihagen.rtt.framework.ITestRunListener;
import de.fernunihagen.rtt.framework.ITestRunListenerRegistry;

/**
 * Test run listener registry.
 * 
 * @author Osama El Hosami
 */
@Contract
public final class TestRunListenerRegistry implements ITestRunListenerRegistry {
	@Invar("null != $target")
	private final Map<String, ITestRunListener> _map;
	@Invar("null != $target")
	private String _id;

	private final static String PROP_ID = "component.name";

	private static final class LoggerHolder {
		private static final Logger _DEFAULT = getRTTCorePlugin().getLogger(
				TestRunListenerRegistry.class.getName());
	}

	@Post("null != $return")
	private Logger getLogger() {
		return LoggerHolder._DEFAULT;
	}

	/**
	 * Reports a status representing an error message.
	 * 
	 * @param message
	 *            Message or null
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>Always</li>
	 *             </ul>
	 */
	@Pre("null != message")
	protected final void reportError(final String message) throws CoreException {
		getRTTCorePlugin().reportError(message);
	}

	/**
	 * Creates a test run listener registry.
	 */
	public TestRunListenerRegistry() {
		_map = Collections
				.synchronizedMap(new HashMap<String, ITestRunListener>());
		_id = getClass().getName().toLowerCase();
	}

	/**
	 * Component activation.
	 * 
	 * @param properties
	 *            Property map
	 */
	protected void activate(final Map<?, ?> properties) {
		_id = (String) properties.get(PROP_ID);

		if (getLogger().isDebugEnabled()) {
			final String message = String.format("[%s (%s) : activated",
					getID(), properties.toString());
			getLogger().debug(message);
		}
	}

	/**
	 * Retrieves the provider identifier.
	 * 
	 * @return Identifier
	 */
	public String getID() {
		return _id;
	}

	/**
	 * Retrieves all registered test run listeners.
	 * 
	 * @return All registered test run listeners
	 */
	@Post("null != $return")
	public final ITestRunListener[] get() {
		return _map.values().toArray(new ITestRunListener[0]);
	}

	/**
	 * Adds a test run listener.
	 * 
	 * @param listener
	 *            Test run listener
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>On misconfiguration</li>
	 *             </ul>
	 */
	@Pre("null != listener")
	public final void addTestRunListener(final ITestRunListener listener)
			throws CoreException {
		final ITestRunListener oldListener = _map.put(listener.getID(),
				listener);
		if (null != oldListener) {
			final String message = String.format(
					"Previously registered listener('%s') '%s' has been replaced"
							+ " by '%s'", listener.getID(), oldListener
							.getName(), listener.getName());
			getLogger().debug(message);
		}
	}

	/**
	 * Removes test run listener.
	 * 
	 * @param listener
	 *            Test run listener
	 */
	@Pre("null != listener")
	public final void removeTestRunListener(final ITestRunListener listener) {
		final String id = listener.getID();
		final ITestRunListener removed = _map.remove(id);
		getLogger().assertLog(removed == listener,
				"Unrequested listener '" + id + "' has been removed");
	}

	/**
	 * Retrieves a test run listener by identifier.
	 * 
	 * @return Test run listener
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If not found</li>
	 *             </ul>
	 */
	@Pre("null != id")
	@Post("null != $return")
	public final ITestRunListener get(final String id) throws CoreException {
		final ITestRunListener listener = _map.get(id);
		if (null == listener) {
			reportError("Test run listener with id '" + id + "' not found");
		}
		return listener;
	}

}
