package de.fernunihagen.rtt.internal.util;

import java.io.File;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

/**
 * Abstract file system node.
 * 
 * @author Osama El Hosami
 */
@Contract
public abstract class AbstractFileNode {
	@Invar("null != path")
	private final File _file;

	/**
	 * Create a file system node.
	 * 
	 * @param file
	 *            File handle
	 */
	@Pre("null != file")
	protected AbstractFileNode(final File file) {
		_file = file;
	}

	/**
	 * Create a file system node.
	 * 
	 * @param location
	 *            File location
	 * 
	 * @return File system node
	 */
	public static AbstractFileNode create(final String location) {
		return create(new File(location));
	}

	/**
	 * Create a file system node.
	 * 
	 * @param file
	 *            File handle
	 * 
	 * @return File system node
	 */
	public static AbstractFileNode create(final File file) {
		if (null == file) {
			throw new NullPointerException("file");
		}
		return file.isFile() ? new FileNode(file) : new DirectoryNode(file);
	}

	/**
	 * Create a file system node.
	 * 
	 * @param dir
	 *            Directory handle
	 * @param name
	 *            File name
	 * 
	 * @return File system node
	 */
	public static AbstractFileNode create(final File dir, final String name) {
		return create(new File(dir, name));
	}

	/**
	 * Retrieves the file handle of file system node
	 * 
	 * @return File handle
	 */
	@Post("null != $return")
	public final File toFile() {
		return _file;
	}

	/**
	 * Retrieves the parent node.
	 * 
	 * @return Parent node
	 */
	@Post("null != $return")
	public final AbstractFileNode getParent() {
		return AbstractFileNode.create(toFile().getParentFile());
	}

	/**
	 * Accepts a file visitor.
	 * 
	 * @param visitor
	 *            File visitor
	 */
	@Pre("null != visitor")
	abstract public void accept(FileNodeVisitor visitor);

	@Override
	public String toString() {
		return toFile().getAbsolutePath();
	}
}
