package de.fernunihagen.rtt.internal.util;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Pre;

/**
 * Decorates a function to make it conditional - functions are only executed in
 * case condition could be satisfied, otherwise null is returned.
 * 
 * @param <R>
 *            Result type
 * @param <X>
 *            Exception type
 * 
 * @author Osama El Hosami
 */
@Contract
public final class ConditionalFunction<R, X extends Throwable> implements
		IDelegate<R, X>, IConditionalCommand<R, X> {
	@Invar("null != $target")
	private final Condition _condition;
	@Invar("null != $target")
	private final IDelegate<R, X> _complied;
	@Invar("null != $target")
	private final IDelegate<R, X> _notComplied;

	@Pre("null != cond && null != complied && null != notComplied")
	private ConditionalFunction(Condition cond, IDelegate<R, X> complied,
			IDelegate<R, X> notComplied) {

		if (complied instanceof ConditionalFunction<?, ?>) {
			final ConditionalFunction<R, X> cfnc = (ConditionalFunction<R, X>) complied;
			complied = cfnc._complied;
			cond = cond.and(cfnc._condition);
		}

		if (notComplied instanceof ConditionalFunction<?, ?>) {
			final ConditionalFunction<R, X> cfnc = (ConditionalFunction<R, X>) notComplied;
			notComplied = cfnc._notComplied;
			cond = cond.and(cfnc._condition);
		}

		_complied = complied;
		_notComplied = notComplied;
		_condition = cond;
	}

	/**
	 * Decorates a function delegate and makes it conditional.
	 * 
	 * @param <R>
	 *            Result type
	 * @param <X>
	 *            Exception type
	 * @param condition
	 *            Condition
	 * @param complied
	 *            Complied delegate
	 * @param notComplied
	 *            Not complied delegate
	 * 
	 * @return The conditional function
	 */
	public static <R, X extends Throwable> ConditionalFunction<R, X> decorate(
			final Condition condition, final IDelegate<R, X> complied,
			final IDelegate<R, X> notComplied) {
		return new ConditionalFunction<R, X>(condition, complied, notComplied);
	}

	@Override
	public R invoke(Object... args) throws X {
		return _condition.invoke(this, args);
	}

	@Override
	@Pre()
	public R invoke(ConditionTrue condition, Object... args) throws X {
		return _complied.invoke(args);
	}

	@Override
	@Pre()
	public R invoke(ConditionFalse condition, Object... args) throws X {
		return _notComplied.invoke(args);
	}
}
