package de.fernunihagen.rtt.internal.util;

import java.io.Serializable;
import java.util.Collection;
import java.util.Map;
import java.util.Set;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Pre;

/**
 * Synchronized cache adapter.
 * 
 * @author Osama El Hosami
 * 
 * @param <K>
 *            Key type
 * @param <V>
 *            Value type
 */
@Contract
public final class SynchronizedCache<K, V> implements ICache<K, V>,
		Serializable {
	private static final long serialVersionUID = 4579612674396312502L;

	@Invar("null != $target")
	private final ICache<K, V> _adaptee;

	/**
	 * Create a synchronized cache
	 * 
	 * @param cache
	 *            Cache
	 */
	@Pre("null != cache")
	public SynchronizedCache(final ICache<K, V> cache) {
		_adaptee = cache;
	}

	@Override
	public int getCapacity() {
		int value;
		synchronized (this) {
			value = _adaptee.getCapacity();
		}
		return value;
	}

	@Override
	public void clear() {
		synchronized (this) {
			_adaptee.clear();
		}
	}

	@Override
	public boolean containsKey(Object arg0) {
		boolean value;
		synchronized (this) {
			value = _adaptee.containsKey(arg0);
		}
		return value;
	}

	@Override
	public boolean containsValue(Object arg0) {
		boolean value;
		synchronized (this) {
			value = _adaptee.containsValue(arg0);
		}
		return value;
	}

	@Override
	public V get(Object arg0) {
		V value;
		synchronized (this) {
			value = _adaptee.get(arg0);
		}
		return value;
	}

	@Override
	public boolean isEmpty() {
		boolean value;
		synchronized (this) {
			value = _adaptee.isEmpty();
		}
		return value;
	}

	@Override
	public V remove(Object arg0) {
		V value;
		synchronized (this) {
			value = _adaptee.remove(arg0);
		}
		return value;
	}

	@Override
	public int size() {
		int value;
		synchronized (this) {
			value = _adaptee.size();
		}
		return value;
	}

	@Override
	public Set<java.util.Map.Entry<K, V>> entrySet() {
		Set<java.util.Map.Entry<K, V>> value;
		synchronized (this) {
			value = _adaptee.entrySet();
		}
		return value;
	}

	@Override
	public Set<K> keySet() {
		Set<K> value;
		synchronized (this) {
			value = _adaptee.keySet();
		}
		return value;
	}

	@Override
	public void putAll(Map<? extends K, ? extends V> arg0) {
		synchronized (this) {
			_adaptee.putAll(arg0);
		}
	}

	@Override
	public Collection<V> values() {
		Collection<V> value;
		synchronized (this) {
			value = _adaptee.values();
		}
		return value;
	}

	@Override
	public V put(K arg0, V arg1) {
		V value;
		synchronized (this) {
			value = _adaptee.put(arg0, arg1);
		}
		return value;
	}

	@Override
	public IFactory<V, ICache<K, V>> getProvider() {
		IFactory<V, ICache<K, V>> value;
		synchronized (this) {
			value = _adaptee.getProvider();
		}
		return value;
	}

	@Override
	public void setProvider(IFactory<V, ICache<K, V>> provider) {
		synchronized (this) {
			_adaptee.setProvider(provider);
		}
	}

	@Override
	public void unsetProvider(IFactory<V, ICache<K, V>> provider) {
		synchronized (this) {
			_adaptee.unsetProvider(provider);
		}
	}
}
